"""Add state_reason to the model.

Revision ID: 229f6f273a56
Revises: 1a44272e8b4c
Create Date: 2016-10-20 10:18:03.019775

"""

# revision identifiers, used by Alembic.
revision = '229f6f273a56'
down_revision = '1a44272e8b4c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('component_builds', sa.Column('state_reason', sa.String(), nullable=True))
    op.add_column('module_builds', sa.Column('state_reason', sa.String(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('module_builds', 'state_reason')
    op.drop_column('component_builds', 'state_reason')
    ### end Alembic commands ###



# Copyright (c) 2017 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from mock import patch, PropertyMock
from gi.repository import GLib

from tests import conf, clean_database
from tests.test_views.test_views import FakeSCM
import module_build_service.messaging
import module_build_service.scheduler.handlers.modules
from module_build_service import build_logs, Modulemd, db
from module_build_service.models import make_session, ModuleBuild, ComponentBuild


class TestModuleInit:

    def setup_method(self, test_method):
        self.fn = module_build_service.scheduler.handlers.modules.init
        self.staged_data_dir = os.path.join(
            os.path.dirname(__file__), '../', 'staged_data')
        testmodule_yml_path = os.path.join(
            self.staged_data_dir, 'testmodule_init.yaml')
        with open(testmodule_yml_path, 'r') as f:
            yaml = f.read()
        scmurl = 'git://pkgs.domain.local/modules/testmodule?#620ec77'
        clean_database()
        with make_session(conf) as session:
            ModuleBuild.create(
                session, conf, 'testmodule', '1', 3, yaml, scmurl, 'mprahl')

    def teardown_method(self, test_method):
        try:
            path = build_logs.path(1)
            os.remove(path)
        except Exception:
            pass

    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder."
           "get_built_rpms_in_module_build")
    @patch('module_build_service.scm.SCM')
    def test_init_basic(self, mocked_scm, built_rpms):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule_init.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        built_rpms.return_value = [
            "foo-0:2.4.48-3.el8+1308+551bfa71",
            "foo-debuginfo-0:2.4.48-3.el8+1308+551bfa71",
            "bar-0:2.5.48-3.el8+1308+551bfa71",
            "bar-debuginfo-0:2.5.48-3.el8+1308+551bfa71",
            "x-0:2.5.48-3.el8+1308+551bfa71",
            "x-debuginfo-0:2.5.48-3.el8+1308+551bfa71"]

        platform_build = ModuleBuild.query.get(1)
        mmd = platform_build.mmd()
        filter_list = Modulemd.SimpleSet()
        filter_list.add("foo")
        filter_list.add("bar")
        mmd.set_rpm_filter(filter_list)
        platform_build.modulemd = mmd.dumps()
        db.session.commit()

        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=2, module_build_state='init')

        with make_session(conf) as session:
            self.fn(config=conf, session=session, msg=msg)
        build = ModuleBuild.query.filter_by(id=2).one()
        # Make sure the module entered the wait state
        assert build.state == 1, build.state
        # Make sure format_mmd was run properly
        xmd_mbs = build.mmd().get_xmd()['mbs']
        assert type(xmd_mbs) is GLib.Variant
        assert xmd_mbs["buildrequires"]["platform"]["filtered_rpms"] == [
            'foo-0:2.4.48-3.el8+1308+551bfa71', 'bar-0:2.5.48-3.el8+1308+551bfa71']

    @patch('module_build_service.scm.SCM')
    def test_init_scm_not_available(self, mocked_scm):
        def mocked_scm_get_latest():
            raise RuntimeError("Failed in mocked_scm_get_latest")

        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        mocked_scm.return_value.get_latest = mocked_scm_get_latest
        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=2, module_build_state='init')
        with make_session(conf) as session:
            self.fn(config=conf, session=session, msg=msg)
        build = ModuleBuild.query.filter_by(id=2).one()
        # Make sure the module entered the failed state
        # since the git server is not available
        assert build.state == 4, build.state

    @patch("module_build_service.config.Config.modules_allow_repository",
           new_callable=PropertyMock, return_value=True)
    @patch('module_build_service.scm.SCM')
    def test_init_includedmodule(self, mocked_scm, mocked_mod_allow_repo):
        FakeSCM(mocked_scm, "includedmodules", ['testmodule_init.yaml'])
        includedmodules_yml_path = os.path.join(
            self.staged_data_dir, 'includedmodules.yaml')
        with open(includedmodules_yml_path, 'r') as f:
            yaml = f.read()
        scmurl = 'git://pkgs.domain.local/modules/includedmodule?#da95886'
        with make_session(conf) as session:
            ModuleBuild.create(
                session, conf, 'includemodule', '1', 3, yaml, scmurl, 'mprahl')
            msg = module_build_service.messaging.MBSModule(
                msg_id=None, module_build_id=3, module_build_state='init')
            self.fn(config=conf, session=session, msg=msg)
        build = ModuleBuild.query.filter_by(id=3).one()
        assert build.state == 1
        assert build.name == 'includemodule'
        batches = {}
        for comp_build in ComponentBuild.query.filter_by(module_id=3).all():
            batches[comp_build.package] = comp_build.batch
        assert batches['perl-List-Compare'] == 2
        assert batches['perl-Tangerine'] == 2
        assert batches['foo'] == 2
        assert batches['tangerine'] == 3
        assert batches['file'] == 4
        # Test that the RPMs are properly merged in xmd
        xmd_rpms = {
            'perl-List-Compare': {'ref': '4f26aeafdb'},
            'perl-Tangerine': {'ref': '4f26aeafdb'},
            'tangerine': {'ref': '4f26aeafdb'},
            'foo': {'ref': '93dea37599'},
            'file': {'ref': 'a2740663f8'},
        }
        assert build.mmd().get_xmd()['mbs']['rpms'] == xmd_rpms

    @patch('module_build_service.models.ModuleBuild.from_module_event')
    @patch('module_build_service.scm.SCM')
    def test_init_when_get_latest_raises(self, mocked_scm, mocked_from_module_event):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '7035bd33614972ac66559ac1fdd019ff6027ad22',
                get_latest_raise=True)
        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=2, module_build_state='init')
        with make_session(conf) as session:
            build = session.query(ModuleBuild).filter_by(id=2).one()
            mocked_from_module_event.return_value = build
            self.fn(config=conf, session=session, msg=msg)
            # Query the database again to make sure the build object is updated
            session.refresh(build)
            # Make sure the module entered the failed state
            assert build.state == 4, build.state
            assert 'Failed to get the latest commit for' in build.state_reason



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
GLib Related helper functions
Originally written by Nils Philippsen in https://pagure.io/modulemd/pull-request/63
"""
from six import text_type
from gi.repository import GLib


def from_variant_dict(d):
    """ Converts a variant dictionary to a Python dictionary
    """
    rv = {}
    for key, value in d.items():
        rv[key] = value.unpack()
    return rv


def variant_str(s):
    """ Converts a string to a GLib.Variant
    """
    if not isinstance(s, str):
        raise TypeError('Only strings are supported for scalars')

    return GLib.Variant('s', s)


def variant_list(l):
    """ Converts a list to a GLib.Variant
    """
    l_variant = list()
    for item in l:
        if item is None:
            item = ''
        if type(item) == str:
            l_variant.append(variant_str(item))
        elif type(item) == text_type:
            l_variant.append(variant_str(item.encode('utf-8')))
        elif type(item) == list:
            l_variant.append(variant_list(item))
        elif type(item) == dict:
            l_variant.append(variant_dict(item))
        elif type(item) == bool:
            l_variant.append(variant_bool(item))
        else:
            raise TypeError('Cannot convert unknown type')
    return GLib.Variant('av', l_variant)


def variant_bool(b):
    """ Converts a boolean to a GLib.Varant
    """
    if not isinstance(b, bool):
        raise TypeError('Only booleans are supported')

    return GLib.Variant('b', b)


def dict_values(d):
    """ Converts each dictionary value to a GLib.Variant
    """
    if not isinstance(d, dict):
        raise TypeError('Only dictionaries are supported for mappings')

    d_variant = dict()
    for k, v in d.items():
        if v is None:
            v = ''
        if type(v) == str:
            d_variant[k] = variant_str(v)
        elif type(v) == text_type:
            d_variant[k] = variant_str(v.encode('utf-8'))
        elif type(v) == list:
            d_variant[k] = variant_list(v)
        elif type(v) == dict:
            d_variant[k] = variant_dict(v)
        elif type(v) == bool:
            d_variant[k] = variant_bool(v)
        else:
            raise TypeError('Cannot convert unknown type')
    return d_variant


def variant_dict(d):
    """ Converts a dictionary to a dictionary of GLib.Variant
    """
    if not isinstance(d, dict):
        raise TypeError('Only dictionaries are supported for mappings')

    d_variant = dict_values(d)
    return GLib.Variant('a{sv}', d_variant)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import re

from module_build_service import conf, log, glib
from module_build_service.resolver import system_resolver


"""
This module handles module stream collision with ursine content before a module
is built.

This kind of collision would happen when packages, which are managed by
Ursa-Major, are used to build a module. Let's see an example, when a module foo
buildrequires bar with stream A, however module bar with another stream B was
already added to ursine content by Ursa-Major when it has been built and moved
to ready state. Hence, MBS has to ensure any packages from module bar:B are not
present in the buildroot.

A technical background:

Generally, each module buildrequires a platform module, which is associated
with ursine content by adding an external repository to the platform module's
tag. That repository is generated from a build tag that inherits from a set
modules' koji tag through its tag inheritance hierarchy. Ursa-Major manages
the inheritance relationship.

Stream collision modules are just those modules added to tag inheritance by
Ursa-Major.
"""


def find_build_tags_from_external_repos(koji_session, repo_infos):
    """Find build tags from external repos

    An external repo added to a tag could be an arbitrary external repository.
    Hence, this method tries best to guess the tags from each external repo's
    URL by a regular expression.

    :param repo_infos: list of mappings represeting external repos information.
    :type repo_infos: list[dict]
    :return: a list of tag names.
    :rtype: list[str]
    """
    re_external_repo_url = r'^{}/repos/(.+-build)/latest/\$arch/?$'.format(
        conf.koji_external_repo_url_prefix.rstrip('/'))
    tag_names = []
    for info in repo_infos:
        match = re.match(re_external_repo_url, info['url'])
        if match:
            name = match.groups()[0]
            if koji_session.getTag(name) is None:
                log.warning('Ignoring the found tag %s because no tag info was found '
                            'with this name.', name)
            else:
                tag_names.append(name)
        else:
            log.warning('The build tag could not be parsed from external repo '
                        '%s whose url is %s.',
                        info['external_repo_name'], info['url'])
    return tag_names


def find_module_koji_tags(koji_session, build_tag):
    """
    Find module koji tags from parents of build tag through the tag inheritance

    MBS supports a few prefixes which are configured in
    ``conf.koij_tag_prefixes``. Tags with a configured prefix will be
    considered as a module's koji tag.

    :param koji_session: instance of Koji client session.
    :type koji_session: ClientSession
    :param str build_tag: tag name, which is the build tag inheriting from
        parent tags where module koji tags are contained.
    :return: list of module koji tags.
    :rtype: list[str]
    """
    return [
        data['name'] for data in koji_session.getFullInheritance(build_tag)
        if any(data['name'].startswith(prefix) for prefix in conf.koji_tag_prefixes)
    ]


def get_modulemds_from_ursine_content(tag):
    """Get all modules metadata which were added to ursine content

    Ursine content is the tag inheritance managed by Ursa-Major by adding
    specific modules' koji_tag.

    Background of module build based on ursine content:

    Each module build buildrequires a platform module, which is a presudo-module
    used to connect to an external repository whose packages will be present
    in the buildroot. In practice, the external repo is generated from a build
    tag which could inherit from a few module koji_tags so that those module's
    RPMs could be build dependencies for some specific packages.

    So, this function is to find out all module koji_tags from the build tag
    and return corresponding module metadata.

    :param str tag: a base module's koji_tag.
    :return: list of module metadata. Empty list will be returned if no ursine
        modules metadata is found.
    :rtype: list[Modulemd.Module]
    """
    from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder
    koji_session = KojiModuleBuilder.get_session(conf, None)
    repos = koji_session.getExternalRepoList(tag)
    build_tags = find_build_tags_from_external_repos(koji_session, repos)
    if not build_tags:
        log.debug('No external repo containing ursine content is found.')
        return []
    modulemds = []
    for tag in build_tags:
        koji_tags = find_module_koji_tags(koji_session, tag)
        for koji_tag in koji_tags:
            md = system_resolver.get_modulemd_by_koji_tag(koji_tag)
            if md:
                modulemds.append(md)
            else:
                log.warning('No module is found by koji_tag \'%s\'', koji_tag)
    return modulemds


def find_stream_collision_modules(buildrequired_modules, koji_tag):
    """
    Find buildrequired modules that are part of the ursine content represented
    by the koji_tag but with a different stream.

    :param dict buildrequired_modules: a mapping of buildrequires, which is just
        the ``xmd/mbs/buildrequires``. This mapping is used to determine if a module
        found from ursine content is a buildrequire with different stream.
    :param str koji_tag: a base module's koji_tag. Modules will be retrieved from
        ursine content associated with this koji_tag and check if there are
        modules that collide.
    :return: a list of NSVC of collision modules. If no collision module is
        found, an empty list is returned.
    :rtype: list[str]
    """
    ursine_modulemds = get_modulemds_from_ursine_content(koji_tag)
    if not ursine_modulemds:
        log.debug('No module metadata is found from ursine content.')
        return []

    collision_modules = [
        item.dup_nsvc()
        for item in ursine_modulemds
        # If some module in the ursine content is one of the buildrequires but has
        # different stream, that is what we want to record here, whose RPMs will be
        # excluded from buildroot by adding them into SRPM module-build-macros as
        # Conflicts.
        if (item.get_name() in buildrequired_modules and
            item.get_stream() != buildrequired_modules[item.get_name()]['stream'])
    ]

    for item in collision_modules:
        name, stream, _ = item.split(':', 2)
        log.info('Buildrequired module %s exists in ursine content with '
                 'different stream %s, whose RPMs will be excluded.',
                 name, stream)

    return collision_modules


def record_stream_collision_modules(mmd):
    """
    Find out modules from ursine content and record those that are buildrequire
    module but have different stream.

    Note that this depends on the result of module stream expansion.

    MBS supports multiple base modules via option conf.base_module_names. A base
    module name could be platform in most cases, but there could be others for
    particular cases in practice. So, each expanded base module stored in
    ``xmd/mbs/buildrequires`` will be handled and will have a new
    key/value pair ``stream_collision_modules: [N-S-V-C, ...]``. This key/value
    will then be handled by the module event handler.

    As a result, a new item is added xmd/mbs/buildrequires/platform/stream_collision_modules,
    which is a list of NSVC strings. Each of them is the module added to ursine
    content by Ursa-Major.

    :param mmd: a module's metadata which will be built.
    :type mmd: Modulemd.Module
    """
    unpacked_xmd = glib.from_variant_dict(mmd.get_xmd())
    buildrequires = unpacked_xmd['mbs']['buildrequires']

    for module_name in conf.base_module_names:
        base_module_info = buildrequires.get(module_name)
        if base_module_info is None:
            log.info(
                'Base module %s is not a buildrequire of module %s. '
                'Skip handling module stream collision for this base module.',
                mmd.get_name())
            continue

        modules_nsvc = find_stream_collision_modules(
            buildrequires, base_module_info['koji_tag'])
        if modules_nsvc:
            base_module_info['stream_collision_modules'] = modules_nsvc
        else:
            log.info('No stream collision module is found against base module %s.',
                     module_name)

    mmd.set_xmd(glib.dict_values(unpacked_xmd))



"""Drops the release column, adds stream column

Revision ID: 145347916a56
Revises: 229f6f273a56
Create Date: 2016-11-10 15:53:21.798038

"""

# revision identifiers, used by Alembic.
revision = '145347916a56'
down_revision = '229f6f273a56'

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('module_builds', schema=None) as batch_op:
        batch_op.add_column(sa.Column('stream', sa.String(), nullable=False))
        batch_op.drop_column('release')

def downgrade():
    with op.batch_alter_table('module_builds', schema=None) as batch_op:
        batch_op.add_column(sa.Column('release', sa.VARCHAR(), nullable=False))
        batch_op.drop_column('stream')



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mock import patch, Mock

from module_build_service import conf, glib
from module_build_service.utils import ursine
from tests import make_module, clean_database


class TestFindModuleKojiTags:
    """Test ursine.find_module_koji_tags"""

    @patch.object(conf, 'koji_tag_prefixes', new=['module'])
    def test_find_out_all_module_koji_tags(self):
        session = Mock()
        session.getFullInheritance.return_value = [
            {'name': 'module-tag1-s-v-c'},
            {'name': 'module-tag2-s-v-c'},
            {'name': 'tag-1'},
        ]

        expected_tags = ['module-tag1-s-v-c', 'module-tag2-s-v-c']

        tags = ursine.find_module_koji_tags(session, 'tag-a-build')
        assert expected_tags == tags

    @patch.object(conf, 'koji_tag_prefixes', new=['module'])
    def test_return_empty_if_no_module_koji_tags(self):
        session = Mock()
        session.getFullInheritance.return_value = [
            {'name': 'tag-1'}, {'name': 'tag-2'},
        ]

        tags = ursine.find_module_koji_tags(session, 'tag-a-build')
        assert [] == tags


class TestFindUrsineRootTags:
    """Test ursine.find_build_tags_from_external_repos"""

    def setup_method(self):
        self.koji_session = Mock()
        self.koji_session.getTag.side_effect = lambda name: \
            None if name == 'X-build' else {'name': name}

    def test_find_build_tags(self):
        with patch.object(conf, 'koji_external_repo_url_prefix',
                          new='http://example.com/brewroot/'):
            tags = ursine.find_build_tags_from_external_repos(self.koji_session, [
                {
                    'external_repo_name': 'tag-1-external-repo',
                    'url': 'http://example.com/brewroot/repos/tag-1-build/latest/$arch/'
                },
                {
                    'external_repo_name': 'tag-2-external-repo',
                    'url': 'http://example.com/brewroot/repos/tag-2-build/latest/$arch/'
                },
            ])

            assert ['tag-1-build', 'tag-2-build'] == tags

    def test_return_emtpy_if_no_match_external_repo_url(self):
        with patch.object(conf, 'koji_external_repo_url_prefix',
                          new='http://example.com/brewroot/'):
            tags = ursine.find_build_tags_from_external_repos(self.koji_session, [
                {
                    'external_repo_name': 'tag-1-external-repo',
                    'url': 'https://another-site.org/repos/tag-1-build/latest/$arch/'
                },
                {
                    'external_repo_name': 'tag-2-external-repo',
                    'url': 'https://another-site.org/repos/tag-2-build/latest/$arch/'
                },
            ])

            assert [] == tags

    def test_some_tag_is_not_koji_tag(self):
        with patch.object(conf, 'koji_external_repo_url_prefix',
                          new='http://example.com/brewroot/'):
            tags = ursine.find_build_tags_from_external_repos(self.koji_session, [
                {
                    'external_repo_name': 'tag-1-external-repo',
                    'url': 'http://example.com/brewroot/repos/tag-1-build/latest/$arch/'
                },
                {
                    'external_repo_name': 'tag-2-external-repo',
                    'url': 'http://example.com/brewroot/repos/X-build/latest/$arch/'
                },
            ])

            assert ['tag-1-build'] == tags


class TestGetModulemdsFromUrsineContent:
    """Test ursine.get_modulemds_from_ursine_content"""

    def setup_method(self):
        clean_database(False)

    def teardown_method(self, test_method):
        clean_database()

    @patch('koji.ClientSession')
    def test_return_empty_if_no_ursine_build_tag_is_found(self, ClientSession):
        session = ClientSession.return_value

        # No module koji_tag in ursine content yet. This will result in empty
        # ursine modulemds is returned.
        session.getFullInheritance.return_value = [
            {'name': 'tag-1.0-build'},
        ]
        session.getExternalRepoList.return_value = [
            {
                'external_repo_name': 'tag-1.0-external-repo',
                'url': 'http://example.com/repos/tag-4-build/latest/$arch/'
            }
        ]

        modulemds = ursine.get_modulemds_from_ursine_content('tag')
        assert [] == modulemds

    @patch.object(conf, 'koji_tag_prefixes', new=['module'])
    @patch('koji.ClientSession')
    def test_get_modulemds(self, ClientSession):
        session = ClientSession.return_value

        # Ensure to to get build tag for further query of ursine content.
        # For this test, the build tag is tag-4-build
        session.getExternalRepoList.return_value = [
            {
                'external_repo_name': 'tag-1.0-external-repo',
                'url': 'http://example.com/repos/tag-4-build/latest/$arch/'
            }
        ]

        # Ensure to return module tags from ursine content of fake build tag
        # specified in above external repo's url.
        def mock_getFullInheritance(tag):
            if tag == 'tag-4-build':
                return [
                    {'name': 'tag-1.0-build'},
                    # Below two modules should be returned and whose modulemd
                    # should be also queried from database.
                    {'name': 'module-name1-s-2020-c'},
                    {'name': 'module-name2-s-2021-c'},
                ]
            raise ValueError('{} is not handled by test.'.format(tag))
        session.getFullInheritance.side_effect = mock_getFullInheritance

        # Defaults to DB resolver, so create fake module builds and store them
        # into database to ensure they can be queried.
        mmd_name1s2020c = make_module(
            'name1:s:2020:c',
            xmd={'mbs': {'koji_tag': 'module-name1-s-2020-c'}})
        mmd_name2s2021c = make_module(
            'name2:s:2021:c',
            xmd={'mbs': {'koji_tag': 'module-name2-s-2021-c'}})

        koji_tag = 'tag'  # It's ok to use arbitrary tag name.
        with patch.object(conf, 'koji_external_repo_url_prefix', new='http://example.com/'):
            modulemds = ursine.get_modulemds_from_ursine_content(koji_tag)

        test_nsvcs = [item.dup_nsvc() for item in modulemds]
        test_nsvcs.sort()

        expected_nsvcs = [mmd_name1s2020c.mmd().dup_nsvc(),
                          mmd_name2s2021c.mmd().dup_nsvc()]
        expected_nsvcs.sort()

        session.getExternalRepoList.assert_called_once_with(koji_tag)
        assert expected_nsvcs == test_nsvcs


class TestRecordStreamCollisionModules:
    """Test ursine.record_stream_collision_modules"""

    @patch.object(conf, 'base_module_names', new=['platform'])
    @patch.object(ursine, 'find_stream_collision_modules')
    def test_nothing_changed_if_no_base_module_is_in_buildrequires(
            self, find_stream_collision_modules):
        xmd = {
            'mbs': {
                'buildrequires': {
                    'modulea': {'stream': 'master'}
                }
            }
        }
        fake_mmd = make_module('name1:s:2020:c', xmd=xmd, store_to_db=False)
        original_xmd = glib.from_variant_dict(fake_mmd.get_xmd())

        with patch.object(ursine, 'log') as log:
            ursine.record_stream_collision_modules(fake_mmd)
            log.info.assert_called_once()
            find_stream_collision_modules.assert_not_called()

        assert original_xmd == glib.from_variant_dict(fake_mmd.get_xmd())

    @patch.object(conf, 'base_module_names', new=['platform'])
    @patch('module_build_service.utils.ursine.get_modulemds_from_ursine_content')
    def test_nothing_changed_if_no_modules_in_ursine_content(
            self, get_modulemds_from_ursine_content):
        xmd = {
            'mbs': {
                'buildrequires': {
                    'modulea': {'stream': 'master'},
                    'platform': {'stream': 'master', 'koji_tag': 'module-rhel-8.0-build'},
                }
            }
        }
        fake_mmd = make_module('name1:s:2020:c', xmd=xmd, store_to_db=False)
        original_xmd = glib.from_variant_dict(fake_mmd.get_xmd())

        get_modulemds_from_ursine_content.return_value = []

        with patch.object(ursine, 'log') as log:
            ursine.record_stream_collision_modules(fake_mmd)
            log.info.assert_called()

        assert original_xmd == glib.from_variant_dict(fake_mmd.get_xmd())

    @patch.object(conf, 'base_module_names', new=['platform', 'project-platform'])
    @patch('module_build_service.utils.ursine.get_modulemds_from_ursine_content')
    def test_add_collision_modules(self, get_modulemds_from_ursine_content):
        xmd = {
            'mbs': {
                'buildrequires': {
                    'modulea': {'stream': 'master'},
                    'foo': {'stream': '1'},
                    'bar': {'stream': '2'},
                    'platform': {'stream': 'master', 'koji_tag': 'module-rhel-8.0-build'},
                    'project-platform': {
                        'stream': 'master', 'koji_tag': 'module-project-1.0-build'
                    },
                }
            }
        }
        fake_mmd = make_module('name1:s:2020:c', xmd=xmd, store_to_db=False)

        def mock_get_ursine_modulemds(koji_tag):
            if koji_tag == 'module-rhel-8.0-build':
                return [
                    make_module('modulea:10:20180813041838:5ea3b708', store_to_db=False),
                    make_module('moduleb:1.0:20180113042038:6ea3b105', store_to_db=False),
                ]
            if koji_tag == 'module-project-1.0-build':
                return [
                    make_module('bar:6:20181013041838:817fa3a8', store_to_db=False),
                    make_module('foo:2:20180113041838:95f078a1', store_to_db=False),
                ]

        get_modulemds_from_ursine_content.side_effect = mock_get_ursine_modulemds

        ursine.record_stream_collision_modules(fake_mmd)

        xmd = glib.from_variant_dict(fake_mmd.get_xmd())
        buildrequires = xmd['mbs']['buildrequires']

        assert (['modulea:10:20180813041838:5ea3b708'] ==
                buildrequires['platform']['stream_collision_modules'])

        modules = buildrequires['project-platform']['stream_collision_modules']
        modules.sort()
        expected_modules = ['bar:6:20181013041838:817fa3a8', 'foo:2:20180113041838:95f078a1']
        assert expected_modules == modules


class TestFindStreamCollisionModules:
    """Test ursine.find_stream_collision_modules"""

    @patch('module_build_service.utils.ursine.get_modulemds_from_ursine_content')
    def test_no_modulemds_found_from_ursine_content(
            self, get_modulemds_from_ursine_content):
        get_modulemds_from_ursine_content.return_value = []
        assert not ursine.find_stream_collision_modules({}, 'koji_tag')

    @patch('module_build_service.utils.ursine.get_modulemds_from_ursine_content')
    def test_no_collisions_found(self, get_modulemds_from_ursine_content):
        xmd_mbs_buildrequires = {
            'modulea': {'stream': 'master'},
            'moduleb': {'stream': '10'},
        }
        get_modulemds_from_ursine_content.return_value = [
            make_module('moduler:1:1:c1', store_to_db=False),
            make_module('modules:2:1:c2', store_to_db=False),
            make_module('modulet:3:1:c3', store_to_db=False),
        ]
        assert [] == ursine.find_stream_collision_modules(
            xmd_mbs_buildrequires, 'koji_tag')

    @patch('module_build_service.utils.ursine.get_modulemds_from_ursine_content')
    def test_collision_modules_are_found(self, get_modulemds_from_ursine_content):
        xmd_mbs_buildrequires = {
            'modulea': {'stream': 'master'},
            'moduleb': {'stream': '10'},
        }
        fake_modules = [
            make_module('moduler:1:1:c1', store_to_db=False),
            make_module('moduleb:6:1:c2', store_to_db=False),
            make_module('modulet:3:1:c3', store_to_db=False),
        ]
        get_modulemds_from_ursine_content.return_value = fake_modules

        assert [fake_modules[1].dup_nsvc()] == \
            ursine.find_stream_collision_modules(
                xmd_mbs_buildrequires, 'koji_tag')



"""Adds the context columns

Revision ID: 9ca1c166f426
Revises: f5b1c5203cce
Create Date: 2018-01-09 16:55:43.416631

"""

# revision identifiers, used by Alembic.
revision = '9ca1c166f426'
down_revision = 'f5b1c5203cce'

from alembic import op
import sqlalchemy as sa

# Data migration imports
from module_build_service import Modulemd
import hashlib
import json
from collections import OrderedDict


modulebuild = sa.Table(
    'module_builds',
    sa.MetaData(),
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('modulemd', sa.String()),
    sa.Column('build_context', sa.String()),
    sa.Column('runtime_context', sa.String()),
)


def upgrade():
    connection = op.get_bind()

    op.add_column('module_builds', sa.Column('build_context', sa.String()))
    op.add_column('module_builds', sa.Column('runtime_context', sa.String()))

    # Determine what the contexts should be based on the modulemd
    for build in connection.execute(modulebuild.select()):
        if not build.modulemd:
            continue
        try:
            mmd = Modulemd.Module().new_from_string(build.modulemd)
            mmd.upgrade()
        except Exception:
            # If the modulemd isn't parseable then skip this build
            continue

        mbs_xmd = mmd.get_xmd().get('mbs', {})
        contexts = {}
        for property_name in ['buildrequires', 'requires']:
            # It's possible this module build was built before MBS filled out xmd or before MBS
            # filled out the requires in xmd. We also have to use keys because GLib.Variant
            # doesn't support `in` directly.
            if property_name not in mbs_xmd.keys():
                break
            mmd_property = getattr(mmd.get_dependencies()[0], 'get_{0}'.format(property_name))()
            if set(mbs_xmd[property_name].keys()) != set(mmd_property.keys()):
                break
            mmd_formatted_property = {
                dep: info['ref'] for dep, info in mbs_xmd[property_name].items()}
            property_json = json.dumps(OrderedDict(sorted(mmd_formatted_property.items())))
            contexts[property_name] = hashlib.sha1(property_json).hexdigest()

        # Update the database now
        if len(contexts) == 2:
            connection.execute(
                modulebuild.update().where(modulebuild.c.id == build.id).values(
                    build_context=contexts['buildrequires'], runtime_context=contexts['requires']))
        elif 'buildrequires' in contexts:
            connection.execute(
                modulebuild.update().where(modulebuild.c.id == build.id).values(
                    build_context=contexts['buildrequires']))
        elif 'requires' in contexts:
            connection.execute(
                modulebuild.update().where(modulebuild.c.id == build.id).values(
                    runtime_context=contexts['requires']))


def downgrade():
    with op.batch_alter_table('module_builds') as b:
        b.drop_column('build_context')
        b.drop_column('runtime_context')



"""Add optional columns for copr

Revision ID: 474697622859
Revises: a1fc0736bca8
Create Date: 2017-02-21 11:18:22.304038

"""

# revision identifiers, used by Alembic.
revision = '474697622859'
down_revision = 'a1fc0736bca8'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('module_builds', sa.Column('copr_owner', sa.String(), nullable=True))
    op.add_column('module_builds', sa.Column('copr_project', sa.String(), nullable=True))


def downgrade():
    op.drop_column('module_builds', 'copr_owner')
    op.drop_column('module_builds', 'copr_project')



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com>

import json
from datetime import datetime

import module_build_service.scm

from mock import patch, PropertyMock, Mock
from shutil import copyfile
from os import path, mkdir
from os.path import dirname
from requests.utils import quote
import hashlib
import pytest

from tests import app, init_data, clean_database, reuse_component_init_data
from tests import read_staged_data
from tests.test_scm import base_dir as scm_base_dir
from module_build_service.errors import UnprocessableEntity
from module_build_service.models import ModuleBuild
from module_build_service import db, version, Modulemd
import module_build_service.config as mbs_config
import module_build_service.scheduler.handlers.modules
from module_build_service.utils import import_mmd, load_mmd


user = ('Homer J. Simpson', set(['packager']))
other_user = ('some_other_user', set(['packager']))
anonymous_user = ('anonymous', set(['packager']))
import_module_user = ('Import M. King', set(['mbs-import-module']))
base_dir = dirname(dirname(__file__))


class FakeSCM(object):
    def __init__(self, mocked_scm, name, mmd_filenames, commit=None, checkout_raise=False,
                 get_latest_raise=False):
        """
        Adds default testing checkout, get_latest and name methods
        to mocked_scm SCM class.

        :param mmd_filenames: List of ModuleMetadata yaml files which
        will be checkouted by the SCM class in the same order as they
        are stored in the list.
        """
        self.mocked_scm = mocked_scm
        self.name = name
        self.commit = commit
        if not isinstance(mmd_filenames, list):
            mmd_filenames = [mmd_filenames]
        self.mmd_filenames = mmd_filenames
        self.checkout_id = 0
        self.sourcedir = None

        if checkout_raise:
            self.mocked_scm.return_value.checkout.side_effect = \
                UnprocessableEntity(
                    "checkout: The requested commit hash was not found within "
                    "the repository. Perhaps you forgot to push. The original "
                    "message was: ")
        else:
            self.mocked_scm.return_value.checkout = self.checkout

        self.mocked_scm.return_value.name = self.name
        self.mocked_scm.return_value.commit = self.commit
        if get_latest_raise:
            self.mocked_scm.return_value.get_latest.side_effect = \
                UnprocessableEntity("Failed to get_latest commit")
        else:
            self.mocked_scm.return_value.get_latest = self.get_latest
        self.mocked_scm.return_value.repository_root = "git://pkgs.stg.fedoraproject.org/modules/"
        self.mocked_scm.return_value.branch = 'master'
        self.mocked_scm.return_value.sourcedir = self.sourcedir
        self.mocked_scm.return_value.get_module_yaml = self.get_module_yaml

    def checkout(self, temp_dir):
        try:
            mmd_filename = self.mmd_filenames[self.checkout_id]
        except Exception:
            mmd_filename = self.mmd_filenames[0]

        self.sourcedir = path.join(temp_dir, self.name)
        mkdir(self.sourcedir)
        base_dir = path.abspath(path.dirname(__file__))
        copyfile(path.join(base_dir, '..', 'staged_data', mmd_filename),
                 self.get_module_yaml())

        self.checkout_id += 1

        return self.sourcedir

    def get_latest(self, ref='master'):
        return hashlib.sha1(ref.encode('utf-8')).hexdigest()[:10]

    def get_module_yaml(self):
        return path.join(self.sourcedir, self.name + ".yaml")


class TestViews:
    def setup_method(self, test_method):
        self.client = app.test_client()
        init_data(2)

    def test_query_build(self):
        rv = self.client.get('/module-build-service/1/module-builds/2')
        data = json.loads(rv.data)
        assert data['id'] == 2
        assert data['context'] == '00000000'
        assert data['name'] == 'nginx'
        assert data['owner'] == 'Moe Szyslak'
        assert data['stream'] == '1'
        assert data['siblings'] == []
        assert data['state'] == 5
        assert data['state_reason'] is None
        assert data['tasks'] == {
            'rpms': {
                'module-build-macros': {
                    'task_id': 12312321,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'module-build-macros-01-1.module+2+b8661ee4',
                },
                'nginx': {
                    'task_id': 12312345,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'nginx-1.10.1-2.module+2+b8661ee4',
                },
            },
        }
        assert data['time_completed'] == '2016-09-03T11:25:32Z'
        assert data['time_modified'] == '2016-09-03T11:25:32Z'
        assert data['time_submitted'] == '2016-09-03T11:23:20Z'
        assert data['rebuild_strategy'] == 'changed-and-after'
        assert data['version'] == '2'

    @pytest.mark.parametrize('api_version', [0, 99])
    def test_query_builds_invalid_api_version(self, api_version):
        rv = self.client.get('/module-build-service/{0}/module-builds/'.format(api_version))
        data = json.loads(rv.data)
        assert data['error'] == 'Not Found'
        assert data['message'] == 'The requested API version is not available'
        assert data['status'] == 404

    def test_query_build_short(self):
        rv = self.client.get('/module-build-service/1/module-builds/2?short=True')
        data = json.loads(rv.data)
        assert data['id'] == 2
        assert data['context'] == '00000000'
        assert data['name'] == 'nginx'
        assert data['state'] == 5
        assert data['state_name'] == 'ready'
        assert data['stream'] == '1'
        assert data['version'] == '2'

    def test_query_build_with_verbose_mode(self):
        rv = self.client.get('/module-build-service/1/module-builds/2?verbose=true')
        data = json.loads(rv.data)
        assert data['base_module_buildrequires'] == []
        assert data['component_builds'] == [1, 2]
        assert data['context'] == '00000000'
        # There is no xmd information on this module, so these values should be None
        assert data['build_context'] is None
        assert data['runtime_context'] is None
        assert data['id'] == 2
        with open(path.join(base_dir, "staged_data", "nginx_mmd.yaml")) as mmd:
            assert data['modulemd'] == mmd.read()
        assert data['name'] == 'nginx'
        assert data['owner'] == 'Moe Szyslak'
        assert data['scmurl'] == ('git://pkgs.domain.local/modules/nginx'
                                  '?#ba95886c7a443b36a9ce31abda1f9bef22f2f8c9')
        assert data['siblings'] == []
        assert data['state'] == 5
        assert data['state_name'] == 'ready'
        assert data['state_reason'] is None
        # State trace is empty because we directly created these builds and didn't have them
        # transition, which creates these entries
        assert data['state_trace'] == []
        assert data['state_url'] == '/module-build-service/1/module-builds/2'
        assert data['stream'] == '1'
        assert data['tasks'] == {
            'rpms': {
                'module-build-macros': {
                    'task_id': 12312321,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'module-build-macros-01-1.module+2+b8661ee4',
                },
                'nginx': {
                    'task_id': 12312345,
                    'state': 1,
                    'state_reason': None,
                    'nvr': 'nginx-1.10.1-2.module+2+b8661ee4',
                },
            },
        }
        assert data['time_completed'] == u'2016-09-03T11:25:32Z'
        assert data['time_modified'] == u'2016-09-03T11:25:32Z'
        assert data['time_submitted'] == u'2016-09-03T11:23:20Z'
        assert data['version'] == '2'
        assert data['rebuild_strategy'] == 'changed-and-after'
        assert data['siblings'] == []

    def test_query_build_with_br_verbose_mode(self):
        reuse_component_init_data()
        rv = self.client.get('/module-build-service/1/module-builds/2?verbose=true')
        data = json.loads(rv.data)
        assert data['base_module_buildrequires'] == [{
            'context': '00000000',
            'id': 1,
            'name': 'platform',
            'state': 5,
            'state_name': 'ready',
            'stream': 'f28',
            'stream_version': 280000,
            'version': '3'
        }]

    def test_pagination_metadata(self):
        rv = self.client.get('/module-build-service/1/module-builds/?per_page=2&page=2')
        meta_data = json.loads(rv.data)['meta']
        assert meta_data['prev'].split('?', 1)[1] in ['per_page=2&page=1', 'page=1&per_page=2']
        assert meta_data['next'].split('?', 1)[1] in ['per_page=2&page=3', 'page=3&per_page=2']
        assert meta_data['last'].split('?', 1)[1] in ['per_page=2&page=4', 'page=4&per_page=2']
        assert meta_data['first'].split('?', 1)[1] in ['per_page=2&page=1', 'page=1&per_page=2']
        assert meta_data['total'] == 7
        assert meta_data['per_page'] == 2
        assert meta_data['pages'] == 4
        assert meta_data['page'] == 2

    def test_pagination_metadata_with_args(self):
        rv = self.client.get('/module-build-service/1/module-builds/?per_page=2&page=2&order_by=id')
        meta_data = json.loads(rv.data)['meta']
        for link in [meta_data['prev'], meta_data['next'], meta_data['last'], meta_data['first']]:
            assert 'order_by=id' in link
            assert 'per_page=2' in link
        assert meta_data['total'] == 7
        assert meta_data['per_page'] == 2
        assert meta_data['pages'] == 4
        assert meta_data['page'] == 2

    def test_query_builds(self):
        rv = self.client.get('/module-build-service/1/module-builds/?per_page=2')
        items = json.loads(rv.data)['items']
        expected = [
            {
                "component_builds": [11, 12],
                "context": "00000000",
                "id": 7,
                "koji_tag": None,
                "name": "testmodule",
                "owner": "some_other_user",
                "rebuild_strategy": "changed-and-after",
                "scmurl": ("git://pkgs.domain.local/modules/testmodule"
                           "?#ca95886c7a443b36a9ce31abda1f9bef22f2f8c9"),
                "siblings": [],
                "state": 1,
                "state_name": "wait",
                "state_reason": None,
                "stream": "4.3.43",
                "tasks": {
                    "rpms": {
                        "module-build-macros": {
                            "nvr": "module-build-macros-01-1.module+7+f95651e2",
                            "state": 1,
                            "state_reason": None,
                            "task_id": 47383994
                        },
                        "rubygem-rails": {
                            "nvr": "postgresql-9.5.3-4.module+7+f95651e2",
                            "state": 3,
                            "state_reason": None,
                            "task_id": 2433434
                        }
                    }
                },
                "time_completed": None,
                "time_modified": "2016-09-03T12:38:40Z",
                "time_submitted": "2016-09-03T12:38:33Z",
                "version": "7",
                "buildrequires": {},
            },
            {
                "component_builds": [9, 10],
                "context": "00000000",
                "id": 6,
                "koji_tag": "module-postgressql-1.2",
                "name": "postgressql",
                "owner": "some_user",
                "rebuild_strategy": "changed-and-after",
                "scmurl": ("git://pkgs.domain.local/modules/postgressql"
                           "?#aa95886c7a443b36a9ce31abda1f9bef22f2f8c9"),
                "siblings": [],
                "state": 3,
                "state_name": "done",
                "state_reason": None,
                "stream": "1",
                "tasks": {
                    "rpms": {
                        "module-build-macros": {
                            "nvr": "module-build-macros-01-1.module+6+fa947d31",
                            "state": 1,
                            "state_reason": None,
                            "task_id": 47383994
                        },
                        "postgresql": {
                            "nvr": "postgresql-9.5.3-4.module+6+fa947d31",
                            "state": 1,
                            "state_reason": None,
                            "task_id": 2433434
                        }
                    }
                },
                "time_completed": "2016-09-03T11:37:19Z",
                "time_modified": "2016-09-03T12:37:19Z",
                "time_submitted": "2016-09-03T12:35:33Z",
                "version": "3",
                "buildrequires": {},
            }
        ]

        assert items == expected

    def test_query_builds_with_context(self):
        clean_database()
        init_data(2, contexts=True)
        rv = self.client.get('/module-build-service/1/module-builds/?context=3a4057d2')
        items = json.loads(rv.data)['items']
        expected = [
            {
                "component_builds": [3, 4],
                "context": "3a4057d2",
                "id": 3,
                "koji_tag": "module-nginx-1.2",
                "name": "nginx",
                "owner": "Moe Szyslak",
                "rebuild_strategy": "changed-and-after",
                "scmurl": ("git://pkgs.domain.local/modules/nginx"
                           "?#ba95886c7a443b36a9ce31abda1f9bef22f2f8c9"),
                "siblings": [2],
                "state": 5,
                "state_name": "ready",
                "state_reason": None,
                "stream": "0",
                "tasks": {
                    "rpms": {
                        "module-build-macros": {
                            "nvr": "module-build-macros-01-1.module+4+0557c87d",
                            "state": 1,
                            "state_reason": None,
                            "task_id": 47383993
                        },
                        "postgresql": {
                            "nvr": "postgresql-9.5.3-4.module+4+0557c87d",
                            "state": 1,
                            "state_reason": None,
                            "task_id": 2433433
                        }
                    }
                },
                "time_completed": "2016-09-03T11:25:32Z",
                "time_modified": "2016-09-03T11:25:32Z",
                "time_submitted": "2016-09-03T11:23:20Z",
                "version": "2",
                "buildrequires": {},
            }
        ]
        assert items == expected

    def test_query_builds_with_id_error(self):
        rv = self.client.get('/module-build-service/1/module-builds/?id=1')
        actual = json.loads(rv.data)
        msg = ('The "id" query option is invalid. Did you mean to go to '
               '"/module-build-service/1/module-builds/1"?')
        expected = {
            'error': 'Bad Request',
            'message': msg,
            "status": 400
        }
        assert actual == expected

    def test_query_builds_with_nsvc(self):
        nsvcs = ["testmodule:4.3.43:7:00000000",
                 "testmodule:4.3.43:7",
                 "testmodule:4.3.43",
                 "testmodule"]

        results = []
        for nsvc in nsvcs:
            rv = self.client.get('/module-build-service/1/module-builds/?nsvc=%s&per_page=2' % nsvc)
            results.append(json.loads(rv.data)['items'])

        nsvc_keys = ["name", "stream", "version", "context"]

        for items, nsvc in zip(results, nsvcs):
            nsvc_parts = nsvc.split(":")
            for item in items:
                for key, part in zip(nsvc_keys, nsvc_parts):
                    assert item[key] == part

    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session")
    def test_query_builds_with_binary_rpm(self, mock_get_session):
        """
        Test for querying MBS with the binary rpm filename. MBS should return all the modules,
        which contain the rpm.
        """
        # update database with builds which contain koji tags.
        reuse_component_init_data()
        mock_rpm_md = {"build_id": 1065871}
        mock_tags = [{"name": "module-testmodule-master-20170219191323-c40c156c"},
                     {"name": "module-testmodule-master-20170219191323-c40c156c-build"},
                     {"name": "non-module-tag"},
                     {"name": "module-testmodule-master-20170109091357-78e4a6fd"}]

        mock_session = Mock()
        mock_session.getRPM.return_value = mock_rpm_md
        mock_session.listTags.return_value = mock_tags
        mock_get_session.return_value = mock_session

        rpm = quote('module-build-macros-0.1-1.testmodule_master_20170303190726.src.rpm')
        rv = self.client.get('/module-build-service/1/module-builds/?rpm=%s' % rpm)
        results = json.loads(rv.data)['items']

        assert len(results) == 2
        assert results[0]["koji_tag"] == "module-testmodule-master-20170219191323-c40c156c"
        assert results[1]["koji_tag"] == "module-testmodule-master-20170109091357-78e4a6fd"

        mock_session.getRPM.assert_called_once_with(
            "module-build-macros-0.1-1.testmodule_master_20170303190726.src.rpm")
        mock_session.listTags.assert_called_once_with(mock_rpm_md["build_id"])

    @patch('module_build_service.config.Config.system',
           new_callable=PropertyMock, return_value="invalid_builder")
    def test_query_builds_with_binary_rpm_not_koji(self, mock_builder):
        rpm = quote('module-build-macros-0.1-1.testmodule_master_20170303190726.src.rpm')
        rv = self.client.get('/module-build-service/1/module-builds/?rpm=%s' % rpm)
        results = json.loads(rv.data)
        expected_error = {
            'error': 'Bad Request',
            'message': 'Configured builder does not allow to search by rpm binary name!',
            'status': 400
        }
        assert rv.status_code == 400
        assert results == expected_error

    def test_query_component_build(self):
        rv = self.client.get('/module-build-service/1/component-builds/1')
        data = json.loads(rv.data)
        assert data['id'] == 1
        assert data['format'] == 'rpms'
        assert data['module_build'] == 2
        assert data['nvr'] == 'nginx-1.10.1-2.module+2+b8661ee4'
        assert data['package'] == 'nginx'
        assert data['state'] == 1
        assert data['state_name'] == 'COMPLETE'
        assert data['state_reason'] is None
        assert data['task_id'] == 12312345

    def test_query_component_build_short(self):
        rv = self.client.get('/module-build-service/1/component-builds/1?short=True')
        data = json.loads(rv.data)
        assert data['id'] == 1
        assert data['format'] == 'rpms'
        assert data['module_build'] == 2
        assert data['nvr'] == 'nginx-1.10.1-2.module+2+b8661ee4'
        assert data['package'] == 'nginx'
        assert data['state'] == 1
        assert data['state_name'] == 'COMPLETE'
        assert data['state_reason'] is None
        assert data['task_id'] == 12312345

    def test_query_component_build_verbose(self):
        rv = self.client.get('/module-build-service/1/component-builds/3?verbose=true')
        data = json.loads(rv.data)
        assert data['id'] == 3
        assert data['format'] == 'rpms'
        assert data['module_build'] == 3
        assert data['nvr'] == 'postgresql-9.5.3-4.module+3+0557c87d'
        assert data['package'] == 'postgresql'
        assert data['state'] == 1
        assert data['state_name'] == 'COMPLETE'
        assert data['state_reason'] is None
        assert data['task_id'] == 2433433
        assert data['state_trace'][0]['reason'] is None
        assert data['state_trace'][0]['time'] is not None
        assert data['state_trace'][0]['state'] == 1
        assert data['state_trace'][0]['state_name'] == 'wait'
        assert data['state_url'], '/module-build-service/1/component-builds/3'

    def test_query_component_builds_filter_format(self):
        rv = self.client.get('/module-build-service/1/component-builds/'
                             '?format=rpms')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 12

    def test_query_component_builds_filter_ref(self):
        rv = self.client.get('/module-build-service/1/component-builds/'
                             '?ref=this-filter-query-should-return-zero-items')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 0

    def test_query_component_builds_filter_tagged(self):
        rv = self.client.get('/module-build-service/1/component-builds/?tagged=true')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 8

    def test_query_component_builds_filter_nvr(self):
        rv = self.client.get('/module-build-service/1/component-builds/?nvr=nginx-1.10.1-2.'
                             'module%2B2%2Bb8661ee4')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 1

    def test_query_component_builds_filter_task_id(self):
        rv = self.client.get('/module-build-service/1/component-builds/?task_id=12312346')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 1

    def test_query_component_builds_filter_state(self):
        rv = self.client.get('/module-build-service/1/component-builds/?state=3')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_component_builds_filter_multiple_states(self):
        rv = self.client.get('/module-build-service/1/component-builds/?state=3&state=1')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 12

    def test_query_builds_filter_name(self):
        rv = self.client.get('/module-build-service/1/module-builds/?name=nginx')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_builds_filter_koji_tag(self):
        rv = self.client.get('/module-build-service/1/module-builds/?koji_tag=module-nginx-1.2')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_builds_filter_completed_before(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?completed_before=2016-09-03T11:30:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_builds_filter_completed_after(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?completed_after=2016-09-03T11:35:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 3

    def test_query_builds_filter_submitted_before(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?submitted_before=2016-09-03T11:35:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_builds_filter_submitted_after(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?submitted_after=2016-09-03T11:35:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 5

    def test_query_builds_filter_modified_before(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?modified_before=2016-09-03T11:35:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 1

    def test_query_builds_filter_modified_after(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?modified_after=2016-09-03T11:35:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 6

    def test_query_builds_filter_owner(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?owner=Moe%20Szyslak')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_builds_filter_state(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?state=3')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2

    def test_query_builds_filter_multiple_states(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?state=3&state=1')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 4

    def test_query_builds_two_filters(self):
        rv = self.client.get('/module-build-service/1/module-builds/?owner=Moe%20Szyslak'
                             '&modified_after=2016-09-03T11:35:00Z')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 1

    def test_query_builds_filter_nsv(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?name=postgressql&stream=1&version=2')
        data = json.loads(rv.data)
        for item in data['items']:
            assert item['name'] == 'postgressql'
            assert item['stream'] == '1'
            assert item['version'] == '2'
        assert data['meta']['total'] == 1

    def test_query_builds_filter_invalid_date(self):
        rv = self.client.get(
            '/module-build-service/1/module-builds/?modified_after=2016-09-03T12:25:00-05:00')
        data = json.loads(rv.data)
        assert data['error'] == 'Bad Request'
        assert data['message'] == ('An invalid Zulu ISO 8601 timestamp was '
                                   'provided for the \"modified_after\" parameter')
        assert data['status'] == 400

    def test_query_builds_order_by(self):
        build = db.session.query(module_build_service.models.ModuleBuild).filter_by(id=2).one()
        build.name = 'candy'
        db.session.add(build)
        db.session.commit()
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_by=name')
        items = json.loads(rv.data)['items']
        assert items[0]['name'] == 'candy'
        assert items[1]['name'] == 'nginx'

    def test_query_builds_order_desc_by(self):
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_desc_by=id')
        items = json.loads(rv.data)['items']
        # Check that the id is items[0]["id"], items[0]["id"] - 1, ...
        for idx, item in enumerate(items):
            assert item["id"] == items[0]["id"] - idx

    def test_query_builds_order_desc_by_context(self):
        clean_database()
        init_data(2, contexts=True)

        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&name=nginx&order_desc_by=context')
        sorted_items = json.loads(rv.data)['items']
        sorted_contexts = [m['context'] for m in sorted_items]

        expected_contexts = ['d5a6c0fa', '795e97c1', '3a4057d2', '10e50d06']
        assert sorted_contexts == expected_contexts

    def test_query_builds_order_by_order_desc_by(self):
        """
        Test that when both order_by and order_desc_by is set,
        we prefer order_desc_by.
        """
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_desc_by=id&order_by=name')
        items = json.loads(rv.data)['items']
        # Check that the id is items[0]["id"], items[0]["id"] - 1, ...
        for idx, item in enumerate(items):
            assert item["id"] == items[0]["id"] - idx

    def test_query_builds_order_by_wrong_key(self):
        rv = self.client.get('/module-build-service/1/module-builds/?'
                             'per_page=10&order_by=unknown')
        data = json.loads(rv.data)
        assert data['status'] == 400
        assert data['error'] == 'Bad Request'
        assert data['message'] == 'An invalid order_by or order_desc_by key was supplied'

    def test_query_base_module_br_filters(self):
        reuse_component_init_data()
        mmd = load_mmd(path.join(base_dir, 'staged_data', 'platform.yaml'), True)
        mmd.set_stream('f30.1.3')
        import_mmd(db.session, mmd)
        platform_f300103 = ModuleBuild.query.filter_by(stream='f30.1.3').one()
        build = ModuleBuild(
            name='testmodule',
            stream='master',
            version=20170109091357,
            state=5,
            build_context='dd4de1c346dcf09ce77d38cd4e75094ec1c08ec3',
            runtime_context='ec4de1c346dcf09ce77d38cd4e75094ec1c08ef7',
            context='7c29193d',
            koji_tag='module-testmodule-master-20170109091357-7c29193d',
            scmurl='git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79',
            batch=3,
            owner='Dr. Pepper',
            time_submitted=datetime(2018, 11, 15, 16, 8, 18),
            time_modified=datetime(2018, 11, 15, 16, 19, 35),
            rebuild_strategy='changed-and-after',
            modulemd=read_staged_data('testmodule'),
        )
        build.buildrequires.append(platform_f300103)
        db.session.add(build)
        db.session.commit()
        # Query by NSVC
        rv = self.client.get(
            '/module-build-service/1/module-builds/?base_module_br=platform:f28:3:00000000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2
        rv = self.client.get(
            '/module-build-service/1/module-builds/?base_module_br=platform:f30.1.3:3:00000000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 1
        # Query by non-existent NVC
        rv = self.client.get(
            '/module-build-service/1/module-builds/?base_module_br=platform:f12:3:00000000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 0
        # Query by name and stream
        rv = self.client.get('/module-build-service/1/module-builds/?base_module_br_name=platform'
                             '&base_module_br_stream=f28')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2
        # Query by stream version
        rv = self.client.get('/module-build-service/1/module-builds/?base_module_br_stream_version='
                             '280000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2
        # Query by lte stream version
        rv = self.client.get('/module-build-service/1/module-builds/?base_module_br_stream_version_'
                             'lte=290000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 2
        # Query by lte stream version with no results
        rv = self.client.get('/module-build-service/1/module-builds/?base_module_br_stream_version_'
                             'lte=270000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 0
        # Query by gte stream version
        rv = self.client.get('/module-build-service/1/module-builds/?base_module_br_stream_version_'
                             'gte=270000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 3
        # Query by gte stream version with no results
        rv = self.client.get('/module-build-service/1/module-builds/?base_module_br_stream_version_'
                             'gte=320000')
        data = json.loads(rv.data)
        assert data['meta']['total'] == 0

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build(self, mocked_scm, mocked_get_user, rscm, api_version):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        post_url = '/module-build-service/{0}/module-builds/'.format(api_version)
        rv = self.client.post(post_url, data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)

        if api_version >= 2:
            assert isinstance(data, list)
            assert len(data) == 1
            data = data[0]

        assert 'component_builds' in data, data
        assert data['component_builds'] == []
        assert data['name'] == 'testmodule'
        assert data['scmurl'] == ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git'
                                  '?#68931c90de214d9d13feefbd35246a81b6cb8d49')
        assert data['version'] == '281'
        assert data['time_submitted'] is not None
        assert data['time_modified'] is not None
        assert data['time_completed'] is None
        assert data['stream'] == 'master'
        assert data['owner'] == 'Homer J. Simpson'
        assert data['id'] == 8
        assert data['rebuild_strategy'] == 'changed-and-after'
        assert data['state_name'] == 'init'
        assert data['state_url'] == '/module-build-service/{0}/module-builds/8'.format(api_version)
        assert len(data['state_trace']) == 1
        assert data['state_trace'][0]['state'] == 0
        assert data['tasks'] == {}
        assert data['siblings'] == []
        mmd = Modulemd.Module().new_from_string(data['modulemd'])
        mmd.upgrade()

        # Make sure the buildrequires entry was created
        module = ModuleBuild.query.get(8)
        assert len(module.buildrequires) == 1
        assert module.buildrequires[0].name == 'platform'
        assert module.buildrequires[0].stream == 'f28'
        assert module.buildrequires[0].version == '3'
        assert module.buildrequires[0].context == '00000000'
        assert module.buildrequires[0].stream_version == 280000

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_no_base_module(self, mocked_scm, mocked_get_user, rscm):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule-no-base-module.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/2/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        assert data == {
            'status': 422,
            'message': ('None of the base module (platform) streams in the buildrequires section '
                        'could be found'),
            'error': 'Unprocessable Entity'
        }

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch('module_build_service.config.Config.rebuild_strategy_allow_override',
           new_callable=PropertyMock, return_value=True)
    def test_submit_build_rebuild_strategy(self, mocked_rmao, mocked_scm, mocked_get_user, hmsc):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'rebuild_strategy': 'only-changed',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        assert data['rebuild_strategy'] == 'only-changed'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch('module_build_service.config.Config.rebuild_strategies_allowed',
           new_callable=PropertyMock, return_value=['all'])
    @patch('module_build_service.config.Config.rebuild_strategy_allow_override',
           new_callable=PropertyMock, return_value=True)
    def test_submit_build_rebuild_strategy_not_allowed(self, mock_rsao, mock_rsa, mocked_scm,
                                                       mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'rebuild_strategy': 'only-changed',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        assert rv.status_code == 400
        expected_error = {
            'error': 'Bad Request',
            'message': ('The rebuild method of "only-changed" is not allowed. Choose from: all.'),
            'status': 400
        }
        assert data == expected_error

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_dep_not_present(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule-no-deps.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        assert rv.status_code == 422
        expected_error = {
            'error': 'Unprocessable Entity',
            'message': 'Cannot find any module builds for chineese_food:good',
            'status': 422
        }
        assert data == expected_error

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_rebuild_strategy_override_not_allowed(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'rebuild_strategy': 'only-changed',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#68931c90de214d9d13feefbd35246a81b6cb8d49')}))
        data = json.loads(rv.data)
        assert rv.status_code == 400
        expected_error = {
            'error': 'Bad Request',
            'message': ('The request contains the "rebuild_strategy" parameter but overriding '
                        'the default isn\'t allowed'),
            'status': 400
        }
        assert data == expected_error

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_componentless_build(self, mocked_scm, mocked_get_user, rscm):
        FakeSCM(mocked_scm, 'fakemodule', 'fakemodule.yaml',
                '3da541559918a808c2402bba5012f6c60b27661c')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)

        assert data['component_builds'] == []
        assert data['name'] == 'fakemodule'
        assert data['scmurl'] == ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git'
                                  '?#68931c90de214d9d13feefbd35246a81b6cb8d49')
        assert data['version'] == '281'
        assert data['time_submitted'] is not None
        assert data['time_modified'] is not None
        assert data['time_completed'] is None
        assert data['stream'] == 'master'
        assert data['owner'] == 'Homer J. Simpson'
        assert data['id'] == 8
        assert data['state_name'] == 'init'
        assert data['rebuild_strategy'] == 'changed-and-after'

    def test_submit_build_auth_error(self):
        base_dir = path.abspath(path.dirname(__file__))
        client_secrets = path.join(base_dir, "client_secrets.json")
        with patch.dict('module_build_service.app.config', {'OIDC_CLIENT_SECRETS': client_secrets}):
            rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
                {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                    'testmodule.git?#48931b90de214d9d13feefbd35246a81b6cb8d49'}))
            data = json.loads(rv.data)
            assert data['message'] == "No 'authorization' header found."
            assert data['status'] == 401
            assert data['error'] == 'Unauthorized'

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_scm_url_error(self, mocked_get_user):
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://badurl.com'}))
        data = json.loads(rv.data)
        assert data['message'] == 'The submitted scmurl git://badurl.com is not allowed'
        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_scm_url_without_hash(self, mocked_get_user):
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git'}))
        data = json.loads(rv.data)
        assert data['message'] == ('The submitted scmurl git://pkgs.stg.fedoraproject.org'
                                   '/modules/testmodule.git is not valid')
        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_bad_modulemd(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, "bad", "bad.yaml")

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        assert data['message'].startswith('The following invalid modulemd was encountered') is True
        assert data['status'] == 422
        assert data['error'] == 'Unprocessable Entity'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_includedmodule_custom_repo_not_allowed(self,
                                                                 mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, "includedmodules", ["includedmodules.yaml",
                                                "testmodule.yaml"])
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)

        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/7',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        assert data['state'] == 4
        assert data['state_reason'] == 'Canceled by some_other_user.'

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build_already_failed(self, mocked_get_user):
        module = ModuleBuild.query.filter_by(id=7).one()
        module.state = 4
        db.session.add(module)
        db.session.commit()
        rv = self.client.patch('/module-build-service/1/module-builds/7',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=('sammy', set()))
    def test_cancel_build_unauthorized_no_groups(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/7',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=('sammy', set(["packager"])))
    def test_cancel_build_unauthorized_not_owner(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/7',
                               data=json.dumps({'state': 'failed'}))
        data = json.loads(rv.data)

        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user',
           return_value=('sammy', set(["packager", "mbs-admin"])))
    def test_cancel_build_admin(self, mocked_get_user):
        with patch("module_build_service.config.Config.admin_groups",
                   new_callable=PropertyMock, return_value=set(["mbs-admin"])):
            rv = self.client.patch('/module-build-service/1/module-builds/7',
                                   data=json.dumps({'state': 'failed'}))
            data = json.loads(rv.data)

            assert data['state'] == 4
            assert data['state_reason'] == 'Canceled by sammy.'

    @patch('module_build_service.auth.get_user',
           return_value=('sammy', set(["packager"])))
    def test_cancel_build_no_admin(self, mocked_get_user):
        with patch("module_build_service.config.Config.admin_groups",
                   new_callable=PropertyMock, return_value=set(["mbs-admin"])):
            rv = self.client.patch('/module-build-service/1/module-builds/7',
                                   data=json.dumps({'state': 'failed'}))
            data = json.loads(rv.data)

            assert data['status'] == 403
            assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build_wrong_param(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/7',
                               data=json.dumps({'some_param': 'value'}))
        data = json.loads(rv.data)

        assert data['status'] == 400
        assert data['error'] == 'Bad Request'
        assert data['message'] == 'Invalid JSON submitted'

    @patch('module_build_service.auth.get_user', return_value=other_user)
    def test_cancel_build_wrong_state(self, mocked_get_user):
        rv = self.client.patch('/module-build-service/1/module-builds/7',
                               data=json.dumps({'state': 'some_state'}))
        data = json.loads(rv.data)

        assert data['status'] == 400
        assert data['error'] == 'Bad Request'
        assert data['message'] == 'The provided state change is not supported'

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_unsupported_scm_scheme(self, mocked_get_user):
        scmurl = 'unsupported://example.com/modules/'
        'testmodule.git?#0000000000000000000000000000000000000000'
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': scmurl}))
        data = json.loads(rv.data)
        assert data['message'] in ("The submitted scmurl {} is not allowed".format(scmurl),
                                   "The submitted scmurl {} is not valid".format(scmurl))
        assert data['status'] == 403
        assert data['error'] == 'Forbidden'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_version_set_error(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule-version-set.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        assert data['status'] == 400
        assert data['message'] == ('The version "123456789" is already defined in the modulemd '
                                   'but it shouldn\'t be since the version is generated based on '
                                   'the commit time')
        assert data['error'] == 'Bad Request'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_wrong_stream(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule-wrong-stream.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49'}))
        data = json.loads(rv.data)
        assert data['status'] == 400
        assert data['message'] == ('The stream "wrong_stream" that is stored in the modulemd '
                                   'does not match the branch "master"')
        assert data['error'] == 'Bad Request'

    @patch('module_build_service.auth.get_user', return_value=user)
    def test_submit_build_set_owner(self, mocked_get_user):
        data = {
            'branch': 'master',
            'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                      'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49',
            'owner': 'foo',
        }
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(data))
        result = json.loads(rv.data)
        assert result['status'] == 400
        assert "The request contains 'owner' parameter" in result['message']

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=anonymous_user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.no_auth", new_callable=PropertyMock,
           return_value=True)
    def test_submit_build_no_auth_set_owner(self, mocked_conf, mocked_scm, mocked_get_user, hmsc):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        data = {
            'branch': 'master',
            'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                      'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49',
            'owner': 'foo',
        }
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(data))
        result = json.loads(rv.data)

        build = ModuleBuild.query.filter(ModuleBuild.id == result['id']).one()
        assert (build.owner == result['owner'] == 'foo') is True

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=anonymous_user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.no_auth", new_callable=PropertyMock)
    def test_patch_set_different_owner(self, mocked_no_auth, mocked_scm, mocked_get_user, hmsc):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        mocked_no_auth.return_value = True
        data = {
            'branch': 'master',
            'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                      'testmodule.git?#68931c90de214d9d13feefbd35246a81b6cb8d49',
            'owner': 'foo',
        }
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(data))
        r1 = json.loads(rv.data)

        url = '/module-build-service/1/module-builds/' + str(r1['id'])
        r2 = self.client.patch(url, data=json.dumps({'state': 'failed'}))
        assert r2.status_code == 403

        r3 = self.client.patch(url, data=json.dumps({'state': 'failed', 'owner': 'foo'}))
        assert r3.status_code == 200

        mocked_no_auth.return_value = False
        r3 = self.client.patch(url, data=json.dumps({'state': 'failed', 'owner': 'foo'}))
        assert r3.status_code == 400
        assert "The request contains 'owner' parameter" in json.loads(r3.data)['message']

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_commit_hash_not_found(self, mocked_scm, mocked_get_user):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '7035bd33614972ac66559ac1fdd019ff6027ad22', checkout_raise=True)

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#7035bd33614972ac66559ac1fdd019ff6027ad22'}))
        data = json.loads(rv.data)
        assert "The requested commit hash was not found within the repository." in data['message']
        assert "Perhaps you forgot to push. The original message was: " in data['message']
        assert data['status'] == 422
        assert data['error'] == 'Unprocessable Entity'

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.allow_custom_scmurls", new_callable=PropertyMock)
    def test_submit_custom_scmurl(self, allow_custom_scmurls, mocked_scm, mocked_get_user, hmsc):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        def submit(scmurl):
            return self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
                                    {'branch': 'master', 'scmurl': scmurl}))

        allow_custom_scmurls.return_value = False
        res1 = submit('git://some.custom.url.org/modules/testmodule.git?#68931c9')
        data = json.loads(res1.data)
        assert data['status'] == 403
        assert data['message'].startswith('The submitted scmurl') is True
        assert data['message'].endswith('is not allowed') is True

        allow_custom_scmurls.return_value = True
        res2 = submit('git://some.custom.url.org/modules/testmodule.git?#68931c9')
        assert res2.status_code == 201

    @pytest.mark.parametrize('br_override_streams, req_override_streams', (
        (['f28'], None),
        (['f28'], ['f28']),
    ))
    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_dep_override(
            self, mocked_scm, mocked_get_user, rscm, br_override_streams, req_override_streams):
        init_data(data_size=1, multiple_stream_versions=True)
        FakeSCM(mocked_scm, 'testmodule', 'testmodule_platform_f290000.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        post_url = '/module-build-service/2/module-builds/'
        scm_url = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#68931c90de214d9d13feef'
                   'bd35246a81b6cb8d49')
        json_input = {
            'branch': 'master',
            'scmurl': scm_url
        }

        if br_override_streams:
            json_input['buildrequire_overrides'] = {'platform': br_override_streams}
            expected_br = set(br_override_streams)
        else:
            expected_br = set(['f29.0.0'])

        if req_override_streams:
            json_input['require_overrides'] = {'platform': req_override_streams}
            expected_req = set(req_override_streams)
        else:
            expected_req = set(['f29.0.0'])

        rv = self.client.post(post_url, data=json.dumps(json_input))
        data = json.loads(rv.data)

        mmd = Modulemd.Module().new_from_string(data[0]['modulemd'])
        assert len(mmd.get_dependencies()) == 1
        dep = mmd.get_dependencies()[0]
        assert set(dep.get_buildrequires()['platform'].get()) == expected_br
        assert set(dep.get_requires()['platform'].get()) == expected_req

    @pytest.mark.parametrize('dep_type', ('buildrequire', 'require'))
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_override_unused(self, mocked_scm, mocked_get_user, dep_type):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule_platform_f290000.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        post_url = '/module-build-service/2/module-builds/'
        scm_url = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#68931c90de214d9d13feef'
                   'bd35246a81b6cb8d49')
        json_input = {
            'branch': 'master',
            'scmurl': scm_url,
        }
        json_input[dep_type + '_overrides'] = {'nonexistent': ['23'], 'nonexistent2': ['2']}

        rv = self.client.post(post_url, data=json.dumps(json_input))
        data = json.loads(rv.data)

        assert data == {
            'error': 'Bad Request',
            'message': ('The {} overrides for the following modules aren\'t applicable: '
                        'nonexistent, nonexistent2').format(dep_type),
            'status': 400
        }

    @pytest.mark.parametrize('optional_params', (
        {'buildrequire_overrides': {'platform': 'f28'}},
        {'buildrequire_overrides': {'platform': 28}},
        {'buildrequire_overrides': 'platform:f28'},
        {'require_overrides': {'platform': 'f28'}},
        {'require_overrides': {'platform': 28}},
        {'require_overrides': 'platform:f28'}
    ))
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_invalid_override(self, mocked_scm, mocked_get_user, optional_params):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule_platform_f290000.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        post_url = '/module-build-service/2/module-builds/'
        scm_url = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#68931c90de214d9d13feef'
                   'bd35246a81b6cb8d49')
        json_input = {
            'branch': 'master',
            'scmurl': scm_url,
        }
        json_input.update(optional_params)

        rv = self.client.post(post_url, data=json.dumps(json_input))
        data = json.loads(rv.data)

        msg = ('The "{}" parameter must be an object with the keys as module names and the values '
               'as arrays of streams')
        if 'buildrequire_overrides' in optional_params:
            msg = msg.format('buildrequire_overrides')
        elif 'require_overrides' in optional_params:
            msg = msg.format('require_overrides')
        assert data == {
            'error': 'Bad Request',
            'message': msg,
            'status': 400
        }

    def test_about(self):
        with patch.object(mbs_config.Config, 'auth_method', new_callable=PropertyMock) as auth:
            auth.return_value = 'kerberos'
            rv = self.client.get('/module-build-service/1/about/')
        data = json.loads(rv.data)
        assert rv.status_code == 200
        assert data == {'auth_method': 'kerberos', 'api_version': 2, 'version': version}

    def test_rebuild_strategy_api(self):
        rv = self.client.get('/module-build-service/1/rebuild-strategies/')
        data = json.loads(rv.data)
        assert rv.status_code == 200
        expected = {
            'items': [
                {
                    'allowed': False,
                    'default': False,
                    'description': 'All components will be rebuilt',
                    'name': 'all'
                },
                {
                    'allowed': True,
                    'default': True,
                    'description': ('All components that have changed and those in subsequent '
                                    'batches will be rebuilt'),
                    'name': 'changed-and-after'
                },
                {
                    'allowed': False,
                    'default': False,
                    'description': 'All changed components will be rebuilt',
                    'name': 'only-changed'
                }
            ]
        }
        assert data == expected

    def test_rebuild_strategy_api_only_changed_default(self):
        with patch.object(mbs_config.Config, 'rebuild_strategy', new_callable=PropertyMock) as r_s:
            r_s.return_value = 'only-changed'
            rv = self.client.get('/module-build-service/1/rebuild-strategies/')
        data = json.loads(rv.data)
        assert rv.status_code == 200
        expected = {
            'items': [
                {
                    'allowed': False,
                    'default': False,
                    'description': 'All components will be rebuilt',
                    'name': 'all'
                },
                {
                    'allowed': False,
                    'default': False,
                    'description': ('All components that have changed and those in subsequent '
                                    'batches will be rebuilt'),
                    'name': 'changed-and-after'
                },
                {
                    'allowed': True,
                    'default': True,
                    'description': 'All changed components will be rebuilt',
                    'name': 'only-changed'
                }
            ]
        }
        assert data == expected

    def test_rebuild_strategy_api_override_allowed(self):
        with patch.object(mbs_config.Config, 'rebuild_strategy_allow_override',
                          new_callable=PropertyMock) as rsao:
            rsao.return_value = True
            rv = self.client.get('/module-build-service/1/rebuild-strategies/')
        data = json.loads(rv.data)
        assert rv.status_code == 200
        expected = {
            'items': [
                {
                    'allowed': True,
                    'default': False,
                    'description': 'All components will be rebuilt',
                    'name': 'all'
                },
                {
                    'allowed': True,
                    'default': True,
                    'description': ('All components that have changed and those in subsequent '
                                    'batches will be rebuilt'),
                    'name': 'changed-and-after'
                },
                {
                    'allowed': True,
                    'default': False,
                    'description': 'All changed components will be rebuilt',
                    'name': 'only-changed'
                }
            ]
        }
        assert data == expected

    def test_cors_header_decorator(self):
        rv = self.client.get('/module-build-service/1/module-builds/')
        assert rv.headers['Access-Control-Allow-Origin'] == '*'

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch.object(module_build_service.config.Config, 'allowed_groups_to_import_module',
                  new_callable=PropertyMock, return_value=set())
    def test_import_build_disabled(self, mocked_groups, mocked_get_user, api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(post_url)
        data = json.loads(rv.data)

        assert data['error'] == 'Forbidden'
        assert data['message'] == (
            'Import module API is disabled.')

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=user)
    def test_import_build_user_not_allowed(self, mocked_get_user, api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(post_url)
        data = json.loads(rv.data)

        assert data['error'] == 'Forbidden'
        assert data['message'] == (
            'Homer J. Simpson is not in any of '
            'set([\'mbs-import-module\']), only set([\'packager\'])')

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    def test_import_build_scm_invalid_json(self, mocked_get_user, api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(post_url, data='')
        data = json.loads(rv.data)

        assert data['error'] == 'Bad Request'
        assert data['message'] == 'Invalid JSON submitted'

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    def test_import_build_scm_url_not_allowed(self, mocked_get_user, api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + '/mariadb'}))
        data = json.loads(rv.data)

        assert data['error'] == 'Forbidden'
        assert data['message'].startswith('The submitted scmurl ')
        assert data['message'].endswith('/tests/scm_data/mariadb is not allowed')

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    @patch.object(module_build_service.config.Config, 'allow_custom_scmurls',
                  new_callable=PropertyMock, return_value=True)
    def test_import_build_scm_url_not_in_list(self, mocked_scmurls, mocked_get_user,
                                              api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + (
                '/mariadb?#b17bea85de2d03558f24d506578abcfcf467e5bc')}))
        data = json.loads(rv.data)

        assert data['error'] == 'Forbidden'
        assert data['message'].endswith(
            '/tests/scm_data/mariadb?#b17bea85de2d03558f24d506578abcfcf467e5bc '
            'is not in the list of allowed SCMs')

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    @patch.object(module_build_service.config.Config, 'scmurls',
                  new_callable=PropertyMock, return_value=['file://'])
    def test_import_build_scm(self, mocked_scmurls, mocked_get_user, api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + (
                '/mariadb?#7cf8fb26db8dbfea075eb5f898cc053139960250')}))
        data = json.loads(rv.data)

        assert 'Module mariadb:10.2:20180724000000:00000000 imported' in data['messages']
        assert data['module']['name'] == 'mariadb'
        assert data['module']['stream'] == '10.2'
        assert data['module']['version'] == '20180724000000'
        assert data['module']['context'] == '00000000'
        assert data['module']['owner'] == 'mbs_import'
        assert data['module']['state'] == 5
        assert data['module']['state_reason'] is None
        assert data['module']['state_name'] == 'ready'
        assert data['module']['scmurl'] is None
        assert data['module']['component_builds'] == []
        assert data['module']['time_submitted'] == data['module']['time_modified'] == \
            data['module']['time_completed']
        assert data['module']['koji_tag'] == 'mariadb-10.2-20180724000000-00000000'
        assert data['module']['siblings'] == []
        assert data['module']['rebuild_strategy'] == 'all'

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    @patch.object(module_build_service.config.Config, 'scmurls',
                  new_callable=PropertyMock, return_value=['file://'])
    def test_import_build_scm_another_commit_hash(self, mocked_scmurls, mocked_get_user,
                                                  api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + (
                '/mariadb?#1a43ea22cd32f235c2f119de1727a37902a49f20')}))
        data = json.loads(rv.data)

        assert 'Module mariadb:10.2:20180724065109:00000000 imported' in data['messages']
        assert data['module']['name'] == 'mariadb'
        assert data['module']['stream'] == '10.2'
        assert data['module']['version'] == '20180724065109'
        assert data['module']['context'] == '00000000'
        assert data['module']['owner'] == 'mbs_import'
        assert data['module']['state'] == 5
        assert data['module']['state_reason'] is None
        assert data['module']['state_name'] == 'ready'
        assert data['module']['scmurl'] is None
        assert data['module']['component_builds'] == []
        assert data['module']['time_submitted'] == data['module']['time_modified'] == \
            data['module']['time_completed']
        assert data['module']['koji_tag'] == 'mariadb-10.2-20180724065109-00000000'
        assert data['module']['siblings'] == []
        assert data['module']['rebuild_strategy'] == 'all'

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    @patch.object(module_build_service.config.Config, 'scmurls',
                  new_callable=PropertyMock, return_value=['file://'])
    def test_import_build_scm_incomplete_nsvc(self, mocked_scmurls, mocked_get_user,
                                              api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + (
                '/mariadb?#b17bea85de2d03558f24d506578abcfcf467e5bc')}))
        data = json.loads(rv.data)

        assert data['error'] == 'Unprocessable Entity'
        assert data['message'] == 'Incomplete NSVC: None:None:0:00000000'

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    @patch.object(module_build_service.config.Config, 'scmurls',
                  new_callable=PropertyMock, return_value=['file://'])
    def test_import_build_scm_yaml_is_bad(self, mocked_scmurls, mocked_get_user,
                                          api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + (
                '/mariadb?#cb7cf7069059141e0797ad2cf5a559fb673ef43d')}))
        data = json.loads(rv.data)

        assert data['error'] == 'Unprocessable Entity'
        assert data['message'].startswith('The following invalid modulemd was encountered')

    @pytest.mark.parametrize('api_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=import_module_user)
    @patch.object(module_build_service.config.Config, 'scmurls',
                  new_callable=PropertyMock, return_value=['file://'])
    def test_import_build_scm_missing_koji_tag(self, mocked_scmurls, mocked_get_user,
                                               api_version):
        post_url = '/module-build-service/{0}/import-module/'.format(api_version)
        rv = self.client.post(
            post_url,
            data=json.dumps({'scmurl': 'file://' + scm_base_dir + (
                '/mariadb?#9ab5fdeba83eb3382413ee8bc06299344ef4477d')}))
        data = json.loads(rv.data)

        assert data['error'] == 'Unprocessable Entity'
        assert data['message'].startswith('\'koji_tag\' is not set in xmd[\'mbs\'] for module')

    def test_buildrequires_is_included_in_json_output(self):
        # Inject xmd/mbs/buildrequires into an existing module build for
        # assertion later.
        from module_build_service.glib import dict_values, from_variant_dict
        from module_build_service.models import make_session
        from module_build_service import conf
        br_modulea = dict(stream='6', version='1', context='1234')
        br_moduleb = dict(stream='10', version='1', context='5678')
        with make_session(conf) as session:
            build = ModuleBuild.query.first()
            mmd = build.mmd()
            xmd = from_variant_dict(mmd.get_xmd())
            mbs = xmd.setdefault('mbs', {})
            buildrequires = mbs.setdefault('buildrequires', {})
            buildrequires['modulea'] = br_modulea
            buildrequires['moduleb'] = br_moduleb
            mmd.set_xmd(dict_values(xmd))
            build.modulemd = mmd.dumps()
            session.commit()

        rv = self.client.get('/module-build-service/1/module-builds/{}'.format(build.id))
        data = json.loads(rv.data)
        buildrequires = data.get('buildrequires', {})

        assert br_modulea == buildrequires.get('modulea')
        assert br_moduleb == buildrequires.get('moduleb')



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

""" Handlers for koji component build events on the message bus. """

import logging

import module_build_service.builder

import koji

from module_build_service import models, log, messaging

logging.basicConfig(level=logging.DEBUG)


def _finalize(config, session, msg, state):
    """ Called whenever a koji build completes or fails. """

    # First, find our ModuleBuild associated with this component, if any.
    component_build = models.ComponentBuild.from_component_event(session, msg)
    try:
        nvr = "{}-{}-{}".format(msg.build_name, msg.build_version,
                                msg.build_release)
    except KeyError:
        nvr = None

    if not component_build:
        log.debug("We have no record of %s" % nvr)
        return

    log.info("Saw relevant component build of %r from %r." % (nvr, msg.msg_id))

    if msg.state_reason:
        state_reason = msg.state_reason
    elif state != koji.BUILD_STATES['COMPLETE']:
        state_reason = "Failed to build artifact %s in Koji" % (msg.build_name)
    else:
        state_reason = ""

    # Mark the state in the db.
    component_build.state = state
    component_build.nvr = nvr
    component_build.state_reason = state_reason
    session.commit()

    parent = component_build.module_build

    # If the macro build failed, then the module is doomed.
    if (component_build.package == 'module-build-macros' and
       state != koji.BUILD_STATES['COMPLETE']):
        parent.transition(config, state=models.BUILD_STATES['failed'],
                          state_reason=state_reason)
        session.commit()
        return

    further_work = []

    # If there are no other components still building in a batch,
    # we can tag all successfully built components in the batch.
    unbuilt_components_in_batch = [
        c for c in parent.current_batch()
        if c.state == koji.BUILD_STATES['BUILDING'] or not c.state
    ]
    if not unbuilt_components_in_batch:
        failed_components_in_batch = [
            c for c in parent.current_batch()
            if (c.state in [koji.BUILD_STATES['FAILED'],
                            koji.BUILD_STATES['CANCELED']])
        ]

        built_components_in_batch = [
            c for c in parent.current_batch()
            if c.state == koji.BUILD_STATES['COMPLETE']
        ]

        builder = module_build_service.builder.GenericBuilder.create_from_module(
            session, parent, config)

        if failed_components_in_batch:
            log.info("Batch done, but not tagging because of failed component builds. Will "
                     "transition the module to \"failed\"")
            parent.transition(config, state=models.BUILD_STATES['failed'],
                              state_reason="Some components failed to build.")
            session.commit()
            return []
        elif not built_components_in_batch:
            # If there are no successfully built components in a batch, there is nothing to tag.
            # The repository won't be regenerated in this case and therefore we generate fake repo
            # change message here.
            log.info("Batch done. No component to tag")
            further_work += [messaging.KojiRepoChange(
                'components::_finalize: fake msg',
                builder.module_build_tag['name'])]
        else:
            built_component_nvrs_in_batch = [c.nvr for c in built_components_in_batch]
            # tag && add to srpm-build group if neccessary
            log.info("Batch done.  Tagging %i component(s) in the build tag." % len(
                built_component_nvrs_in_batch))
            log.debug("%r" % built_component_nvrs_in_batch)
            # TODO: install=component_build.build_time_only works here because module-build-macros
            # is alone in its batch and the only component with build_time_only set. All other
            # batches will have install=False. If we expand to have batches with a mix of
            # build_time_only and not components, then this logic will need to change.
            builder.buildroot_add_artifacts(
                built_component_nvrs_in_batch, install=component_build.build_time_only)

            # Do not tag packages which only belong to the build tag to the dest tag
            component_nvrs_to_tag_in_dest = [c.nvr for c in built_components_in_batch
                                             if c.build_time_only is False]
            log.info("Tagging %i component(s) in the dest tag." % len(
                component_nvrs_to_tag_in_dest))
            if component_nvrs_to_tag_in_dest:
                builder.tag_artifacts(component_nvrs_to_tag_in_dest)

        session.commit()
    elif (any([c.state != koji.BUILD_STATES['BUILDING']
          for c in unbuilt_components_in_batch])):
        # We are not in the middle of the batch building and
        # we have some unbuilt components in this batch. We might hit the
        # concurrent builds threshold in previous call of continue_batch_build
        # done in repos.py:done(...), but because we have just finished one
        # build, try to call continue_batch_build again so in case we hit the
        # threshold previously, we will submit another build from this batch.
        builder = module_build_service.builder.GenericBuilder.create_from_module(
            session, parent, config)
        further_work += module_build_service.utils.continue_batch_build(
            config, parent, session, builder)
    return further_work


def complete(config, session, msg):
    return _finalize(config, session, msg, state=koji.BUILD_STATES['COMPLETE'])


def failed(config, session, msg):
    return _finalize(config, session, msg, state=koji.BUILD_STATES['FAILED'])


def canceled(config, session, msg):
    return _finalize(config, session, msg, state=koji.BUILD_STATES['CANCELED'])



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluža <jkaluza@redhat.com>


import logging
import os
import koji
import kobo.rpmlib
import pipes
import platform
import re
import subprocess
import threading

from module_build_service import conf, log, Modulemd
import module_build_service.scm
import module_build_service.utils
import module_build_service.scheduler
import module_build_service.scheduler.consumer

from module_build_service.builder.base import GenericBuilder
from module_build_service.builder.utils import (
    create_local_repo_from_koji_tag,
    execute_cmd,
    find_srpm,
    get_koji_config
)
from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder

from module_build_service import models

logging.basicConfig(level=logging.DEBUG)


class MockModuleBuilder(GenericBuilder):
    backend = "mock"
    # Global build_id/task_id we increment when new build is executed.
    _build_id_lock = threading.Lock()
    _build_id = 1
    _config_lock = threading.Lock()

    # Load mock config file template
    for cf in conf.mock_config_file:
        try:
            with open(cf) as f:
                mock_config_template = f.read()
            break
        except IOError:
            pass
    else:
        raise IOError("None of {} mock config files found."
                      .format(conf.mock_config_file))

    # Load yum config file template
    for cf in conf.yum_config_file:
        try:
            with open(cf) as f:
                yum_config_template = f.read()
            break
        except IOError:
            pass
    else:
        raise IOError("None of {} yum config files found."
                      .format(conf.yum_config_file))

    @module_build_service.utils.validate_koji_tag('tag_name')
    def __init__(self, owner, module, config, tag_name, components):
        self.module_str = module.name
        self.module = module
        self.tag_name = tag_name
        self.config = config
        self.groups = []
        self.yum_conf = MockModuleBuilder.yum_config_template
        self.koji_session = None

        # Auto-detect arch (if possible) or fallback to the configured one
        if conf.arch_autodetect:
            arch_detected = platform.machine()
            if arch_detected:
                self.arch = arch_detected
            else:
                log.warning("Couldn't determine machine arch. Falling back "
                            "to configured arch.")
                self.arch = conf.arch_fallback
        else:
            self.arch = conf.arch_fallback
        log.info("Machine arch setting: {}".format(self.arch))

        # Create main directory for this tag
        self.tag_dir = os.path.join(self.config.mock_resultsdir, tag_name)
        if not os.path.exists(self.tag_dir):
            os.makedirs(self.tag_dir)

        # Create "results" sub-directory for this tag to store build results
        # and local repository.
        self.resultsdir = os.path.join(self.tag_dir, "results")
        if not os.path.exists(self.resultsdir):
            os.makedirs(self.resultsdir)

        # Create "config" sub-directory.
        self.configdir = os.path.join(self.tag_dir, "config")
        if not os.path.exists(self.configdir):
            os.makedirs(self.configdir)

        # Generate path to mock config and add local repository there.
        self._add_repo("localrepo", "file://" + self.resultsdir, "metadata_expire=1\n")

        # Remove old files from the previous build of this tag but only
        # before the first build is done, otherwise we would remove files
        # which we already build in this module build.
        if MockModuleBuilder._build_id == 1:
            # Remove all RPMs from the results directory, but keep old logs.
            for name in os.listdir(self.resultsdir):
                if name.endswith(".rpm"):
                    os.remove(os.path.join(self.resultsdir, name))

            # Remove the old RPM repository from the results directory.
            if os.path.exists(os.path.join(self.resultsdir, "repodata/repomd.xml")):
                os.remove(os.path.join(self.resultsdir, "repodata/repomd.xml"))

            # Remove old config files from config directory.
            for name in os.listdir(self.configdir):
                os.remove(os.path.join(self.configdir, name))

        log.info("MockModuleBuilder initialized, tag_name=%s, tag_dir=%s" %
                 (tag_name, self.tag_dir))

    @property
    def module_build_tag(self):
        # Workaround koji specific code in modules.py
        return {"name": self.tag_name}

    def _createrepo(self, include_module_yaml=False):
        """
        Creates the repository using "createrepo_c" command in the resultsdir.
        """
        log.debug("Creating repository in %s" % self.resultsdir)
        path = self.resultsdir
        repodata_path = os.path.join(path, "repodata")

        # Remove old repodata files
        if os.path.exists(repodata_path):
            for name in os.listdir(repodata_path):
                os.remove(os.path.join(repodata_path, name))

        # We pass an explicit package list to createrepo_c, otherwise, it will
        # walk the target directory recursively, instead of just finding the
        # files at the toplevel.
        pkglist = os.path.join(path, "pkglist")
        pkglist_f = open(pkglist, "w")

        # Generate the mmd the same way as pungi does.
        m1 = models.ModuleBuild.query.filter(models.ModuleBuild.name == self.module_str).one()
        m1_mmd = m1.mmd()
        artifacts = Modulemd.SimpleSet()

        rpm_files = [f
                     for f in os.listdir(self.resultsdir)
                     if f.endswith(".rpm")]

        if rpm_files:
            output = subprocess.check_output(['rpm',
                                              '--queryformat',
                                              '%{NAME} %{EPOCHNUM} %{VERSION} %{RELEASE} %{ARCH}\n',
                                              '-qp'] + rpm_files,
                                             cwd=self.resultsdir,
                                             universal_newlines=True)
            nevras = output.strip().split('\n')
            if len(nevras) != len(rpm_files):
                raise RuntimeError("rpm -qp returned an unexpected number of lines")

            for rpm_file, nevra in zip(rpm_files, nevras):
                name, epoch, version, release, arch = nevra.split()

                if m1.last_batch_id() == m1.batch:
                    # If RPM is filtered-out, do not add it to artifacts list.
                    if name in m1_mmd.get_rpm_filter().get():
                        continue

                pkglist_f.write(rpm_file + '\n')
                artifacts.add('{}-{}:{}-{}.{}'.format(name, epoch, version, release, arch))

        pkglist_f.close()
        m1_mmd.set_rpm_artifacts(artifacts)

        # Generate repo.
        execute_cmd(['/usr/bin/createrepo_c', '--pkglist', pkglist, path])

        # ...and inject modules.yaml there if asked.
        if include_module_yaml:
            mmd_path = os.path.join(path, "modules.yaml")
            m1_mmd.dump(mmd_path)
            execute_cmd(['/usr/bin/modifyrepo_c', '--mdtype=modules', mmd_path, repodata_path])

    def _add_repo(self, name, baseurl, extra=""):
        """
        Adds repository to Mock config file. Call _write_mock_config() to
        actually write the config file to filesystem.
        """
        self.yum_conf += "[%s]\n" % name
        self.yum_conf += "name=%s\n" % name
        self.yum_conf += "baseurl=%s\n" % baseurl
        self.yum_conf += extra
        self.yum_conf += "enabled=1\n\n"

    def _load_mock_config(self):
        """
        Loads the variables which are generated only during the first
        initialization of mock config. This should be called before
        every _write_mock_config otherwise we overwrite Mock
        repositories or groups ...
        """

        # We do not want to load old file from previous builds here, so if
        # this is the first build in this module, skip the load completely.
        if MockModuleBuilder._build_id == 1:
            return

        with MockModuleBuilder._config_lock:
            infile = os.path.join(self.configdir, "mock.cfg")
            with open(infile, 'r') as f:
                # This looks scary, but it is the way how mock itself loads the
                # config file ...
                config_opts = {}
                code = compile(f.read(), infile, 'exec')
                # pylint: disable=exec-used
                exec(code)

                self.groups = config_opts["chroot_setup_cmd"].split(" ")[1:]
                self.yum_conf = config_opts['yum.conf']

    def _write_mock_config(self):
        """
        Writes Mock config file to local file.
        """

        with MockModuleBuilder._config_lock:
            config = str(MockModuleBuilder.mock_config_template)
            config = config.replace("$root", "%s-%s" % (self.tag_name,
                                                        str(threading.current_thread().name)))
            config = config.replace("$arch", self.arch)
            config = config.replace("$group", " ".join(self.groups))
            config = config.replace("$yum_conf", self.yum_conf)

            # We write the most recent config to "mock.cfg", so thread-related
            # configs can be later (re-)generated from it using _load_mock_config.
            outfile = os.path.join(self.configdir, "mock.cfg")
            with open(outfile, 'w') as f:
                f.write(config)

            # Write the config to thread-related configuration file.
            outfile = os.path.join(self.configdir, "mock-%s.cfg" %
                                   str(threading.current_thread().name))
            with open(outfile, 'w') as f:
                f.write(config)

    def buildroot_connect(self, groups):
        self._load_mock_config()
        self.groups = list(set().union(groups["build"], self.groups))
        log.debug("Mock builder groups: %s" % self.groups)
        self._write_mock_config()

    def buildroot_prep(self):
        pass

    def buildroot_resume(self):
        pass

    def buildroot_ready(self, artifacts=None):
        return True

    def buildroot_add_dependency(self, dependencies):
        pass

    def buildroot_add_artifacts(self, artifacts, install=False):
        self._createrepo()

        # TODO: This is just hack to install module-build-macros into the
        # buildroot. We should really install the RPMs belonging to the
        # right source RPM into the buildroot here, but we do not track
        # what RPMs are output of particular SRPM build yet.
        for artifact in artifacts:
            if artifact and artifact.startswith("module-build-macros"):
                self._load_mock_config()
                self.groups.append("module-build-macros")
                self._write_mock_config()

        from module_build_service.scheduler.consumer import fake_repo_done_message
        fake_repo_done_message(self.tag_name)

    def tag_artifacts(self, artifacts):
        pass

    def buildroot_add_repos(self, dependencies):
        self._load_mock_config()
        for tag in dependencies:
            # If tag starts with mock_resultdir, it means it is path to local
            # module build repository.
            if tag.startswith(conf.mock_resultsdir):
                repo_name = os.path.basename(tag)
                if repo_name.startswith("module-"):
                    repo_name = repo_name[7:]
                repo_dir = tag
                baseurl = "file://" + repo_dir
            else:
                repo_name = tag
                koji_config = get_koji_config(self.config)
                koji_session = koji.ClientSession(koji_config.server, opts=koji_config)
                repo = koji_session.getRepo(repo_name)
                if repo:
                    baseurl = koji.PathInfo(topdir=koji_config.topurl).repo(repo["id"], repo_name)
                    baseurl = '{0}/{1}/'.format(baseurl, self.arch)
                else:
                    repo_dir = os.path.join(self.config.cache_dir, "koji_tags", tag)
                    create_local_repo_from_koji_tag(self.config, tag, repo_dir,
                                                    [self.arch, "noarch"])
                    baseurl = "file://" + repo_dir
                # Check to see if there are any external repos tied to the tag
                for ext_repo in koji_session.getTagExternalRepos(repo_name):
                    self._add_repo(ext_repo["external_repo_name"], ext_repo["url"])
            self._add_repo(repo_name, baseurl)
        self._write_mock_config()

    def _send_build_change(self, state, source, build_id):
        try:
            nvr = kobo.rpmlib.parse_nvr(source)
        except ValueError:
            nvr = {"name": source, "release": "unknown", "version": "unknown"}

        # build_id=1 and task_id=1 are OK here, because we are building just
        # one RPM at the time.
        msg = module_build_service.messaging.KojiBuildChange(
            msg_id='a faked internal message',
            build_id=build_id,
            task_id=build_id,
            build_name=nvr["name"],
            build_new_state=state,
            build_release=nvr["release"],
            build_version=nvr["version"]
        )
        module_build_service.scheduler.consumer.work_queue_put(msg)

    def _save_log(self, resultsdir, log_name, artifact_name):
        old_log = os.path.join(resultsdir, log_name)
        new_log = os.path.join(resultsdir, artifact_name + "-" + log_name)
        if os.path.exists(old_log):
            os.rename(old_log, new_log)

    def _purge_useless_logs(self):
        """
        Remove empty or otherwise useless log files
        """
        for logf in os.listdir(self.resultsdir):

            log_path = os.path.join(self.resultsdir, logf)

            # Remove empty files
            if os.path.isfile(log_path) and os.path.getsize(log_path) == 0:
                os.remove(log_path)

            # Remove other files containing useless information
            elif logf.endswith('-srpm-stdout.log'):
                with open(log_path) as f:
                    data = f.read(4096)
                    if re.match("Downloading [^\n]*\n\n\nWrote: [^\n]", data):
                        os.remove(log_path)

    def build_srpm(self, artifact_name, source, build_id, builder):
        """
        Builds the artifact from the SRPM.
        """
        state = koji.BUILD_STATES['BUILDING']

        # Use the mock config associated with this thread.
        mock_config = os.path.join(self.configdir,
                                   "mock-%s.cfg" % str(threading.current_thread().name))

        # Open the logs to which we will forward mock stdout/stderr.
        mock_stdout_log = open(os.path.join(self.resultsdir,
                                            artifact_name + "-mock-stdout.log"), "w")
        mock_stderr_log = open(os.path.join(self.resultsdir,
                                            artifact_name + "-mock-stderr.log"), "w")

        srpm = artifact_name
        resultsdir = builder.resultsdir
        try:
            # Initialize mock.
            execute_cmd(["mock", "-v", "-r", mock_config, "--init"],
                        stdout=mock_stdout_log, stderr=mock_stderr_log)

            # Start the build and store results to resultsdir
            builder.build(mock_stdout_log, mock_stderr_log)
            srpm = find_srpm(resultsdir)

            # Emit messages simulating complete build. These messages
            # are put in the scheduler's work queue and are handled
            # by MBS after the build_srpm() method returns and scope gets
            # back to scheduler.main.main() method.
            state = koji.BUILD_STATES['COMPLETE']
            self._send_build_change(state, srpm, build_id)

            with open(os.path.join(resultsdir, "status.log"), 'w') as f:
                f.write("complete\n")
        except Exception as e:
            log.error("Error while building artifact %s: %s" % (artifact_name,
                                                                str(e)))

            # Emit messages simulating complete build. These messages
            # are put in the scheduler's work queue and are handled
            # by MBS after the build_srpm() method returns and scope gets
            # back to scheduler.main.main() method.
            state = koji.BUILD_STATES['FAILED']
            self._send_build_change(state, srpm,
                                    build_id)
            with open(os.path.join(resultsdir, "status.log"), 'w') as f:
                f.write("failed\n")

        mock_stdout_log.close()
        mock_stderr_log.close()

        self._save_log(resultsdir, "state.log", artifact_name)
        self._save_log(resultsdir, "root.log", artifact_name)
        self._save_log(resultsdir, "build.log", artifact_name)
        self._save_log(resultsdir, "status.log", artifact_name)

        # Copy files from thread-related resultsdire to the main resultsdir.
        for name in os.listdir(resultsdir):
            os.rename(os.path.join(resultsdir, name), os.path.join(self.resultsdir, name))

        # Depending on the configuration settings, remove/keep useless log files
        if conf.mock_purge_useless_logs:
            self._purge_useless_logs()

        # We return BUILDING state here even when we know it is already
        # completed or failed, because otherwise utils.start_build_batch
        # would think this component is already built and also tagged, but
        # we have just built it - tagging will happen as result of build
        # change message we are sending above using _send_build_change.
        # It is just to make this backend compatible with other backends,
        # which return COMPLETE here only in case the resulting build is
        # already in repository ready to be used. This is not a case for Mock
        # backend in the time we return here.
        reason = "Building %s in Mock" % (artifact_name)
        return build_id, koji.BUILD_STATES['BUILDING'], reason, None

    def build(self, artifact_name, source):
        log.info("Starting building artifact %s: %s" % (artifact_name, source))

        # Load global mock config for this module build from mock.cfg and
        # generate the thread-specific mock config by writing it to fs again.
        self._load_mock_config()
        self._write_mock_config()
        mock_config = os.path.join(self.configdir, "mock-%s.cfg"
                                   % str(threading.current_thread().name))

        # Get the build-id in thread-safe manner.
        build_id = None
        with MockModuleBuilder._build_id_lock:
            MockModuleBuilder._build_id += 1
            build_id = int(MockModuleBuilder._build_id)

        # Clear resultsdir associated with this thread or in case it does not
        # exist, create it.
        resultsdir = os.path.join(self.resultsdir,
                                  str(threading.current_thread().name))
        if os.path.exists(resultsdir):
            for name in os.listdir(resultsdir):
                os.remove(os.path.join(resultsdir, name))
        else:
            os.makedirs(resultsdir)

        if source.endswith('.src.rpm'):
            builder = SRPMBuilder(mock_config, resultsdir, source)
        else:
            # Otherwise, assume we're building from some scm repo
            builder = SCMBuilder(mock_config, resultsdir, source, artifact_name)
        return self.build_srpm(artifact_name, source, build_id, builder)

    @staticmethod
    def get_disttag_srpm(disttag, module_build):
        # @FIXME
        return KojiModuleBuilder.get_disttag_srpm(disttag, module_build)

    def cancel_build(self, task_id):
        pass

    def list_tasks_for_components(self, component_builds=None, state='active'):
        pass

    def repo_from_tag(cls, config, tag_name, arch):
        pass

    def finalize(self):
        # If the state is "done", run one last createrepo, to include
        # the module metadata. We don't want to do this for failed builds,
        # since that makes it impossible to retry a build manually.
        if self.module.state == models.BUILD_STATES["done"]:
            self._createrepo(include_module_yaml=True)


class BaseBuilder(object):
    def __init__(self, config, resultsdir):
        self.config = config
        self.resultsdir = resultsdir
        self.cmd = ["mock", "-v", "-r", config,
                    "--no-clean",
                    "--resultdir=%s" % resultsdir]

    def build(self, stdout, stderr):
        execute_cmd(self.cmd, stdout=stdout, stderr=stderr)


class SRPMBuilder(BaseBuilder):
    def __init__(self, config, resultsdir, source):
        super(SRPMBuilder, self).__init__(config, resultsdir)
        self.cmd.extend(["--rebuild", source])


class SCMBuilder(BaseBuilder):
    def __init__(self, config, resultsdir, source, artifact_name):
        super(SCMBuilder, self).__init__(config, resultsdir)
        with open(config, "a") as f:
            branch = source.split("?#")[1]
            distgit_cmds = self._get_distgit_commands(source)
            distgit_get = distgit_cmds[0].format(artifact_name)

            # mock-scm cannot checkout particular commit hash, but only branch.
            # We therefore use a command that combines the distgit-command with
            # checking out a particular commit hash.
            # See https://bugzilla.redhat.com/show_bug.cgi?id=1459437 for
            # more info. Once mock-scm supports this feature, we can remove
            # this code.
            distgit_get_branch = \
                "sh -c {}'; git -C {} checkout {}'".format(pipes.quote(distgit_get),
                                                           artifact_name,
                                                           branch)

            f.writelines([
                "config_opts['scm'] = True\n",
                "config_opts['scm_opts']['method'] = 'distgit'\n",
                "config_opts['scm_opts']['package'] = '{}'\n".format(
                    artifact_name),
                "config_opts['scm_opts']['distgit_get'] = {!r}\n".format(
                    distgit_get_branch),
                "config_opts['scm_opts']['distgit_src_get'] = '{}'\n".format(
                    distgit_cmds[1]),
            ])

    def _make_executable(self, path):
        mode = os.stat(path).st_mode
        mode |= (mode & 0o444) >> 2    # copy R bits to X
        os.chmod(path, mode)

    def _get_distgit_commands(self, source):
        for host, cmds in conf.distgits.items():
            if source.startswith(host):
                return cmds
        raise KeyError("No defined commands for {}".format(source))

    def get_average_build_time(self, component):
        """
        Get the average build time of the component from Koji
        :param component: a ComponentBuild object
        :return: a float of the average build time in seconds
        """
        # We currently don't track build times in MBS directly, so we can use Koji to get a decent
        # estimate
        if not self.koji_session:
            # If Koji is not configured on the system, then just return 0.0 for components
            try:
                self.koji_session = KojiModuleBuilder.get_session(self.config, self.owner)
                # If the component has not been built before, then None is returned. Instead,
                # let's return 0.0 so the type is consistent
                return self.koji_session.getAverageBuildDuration(component.package) or 0.0
            except Exception:
                log.debug('The Koji call to getAverageBuildDuration failed. Is Koji properly '
                          'configured?')
                return 0.0
